------------------------------------------------------------------------------
--                                                                          --
--                An Ada2005 binding to the Qt C++ framework                --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2009-2011 Vadim Godunko <vgodunko@gmail.com>                 --
--                                                                          --
-- QtAda is free software;  you can  redistribute it and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. QtAda is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License distributed with QtAda;  see file COPYING.  If not, write --
-- to  the  Free Software Foundation,  51  Franklin  Street,  Fifth  Floor, --
-- Boston, MA 02110-1301, USA.                                              --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 3812 $ $Date: 2011-01-07 16:15:27 +0300 (Fri, 07 Jan 2011) $
------------------------------------------------------------------------------
with Ada.Containers.Hashed_Maps;
with Ada.Strings.Wide_Unbounded;

with Asis.Compilation_Units;
with Asis.Declarations;
with Asis.Definitions;
with Asis.Elements;
with Asis.Expressions;

with Amoc.Asis_Implementation;
with Amoc.Diagnoses;
with Amoc.Model;

package body Amoc.Analizer.Types is

   type Qt_Type_Info is record
      Kind        : Amoc.Model.Type_Kinds;
      Qt_Name     : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Ada_Package : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Ada_Name    : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
   end record;

   function "+" (Right : Wide_String)
     return Ada.Strings.Wide_Unbounded.Unbounded_Wide_String
       renames Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String;

   function "-" (Right : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String)
     return Wide_String
       renames Ada.Strings.Wide_Unbounded.To_Wide_String;

   procedure Process_Qt_Type (Info : Qt_Type_Info);

   function Is_Derived
    (Declaration      : Asis.Element;
     Type_Declaration : Asis.Element;
     Only_Public      : Boolean)
       return Boolean;
   --  Returns True if the type specified by Declaration derived from
   --  the type specified by Type_Declaration. If Only_Public is True then
   --  analyze only public declarations and skip private declarations.

   function Is_Child_Unit
    (Unit   : Asis.Compilation_Unit;
     Parent : Asis.Compilation_Unit)
       return Boolean;
   --  Returns True if Unit is direct on indirect child of Parent unit.

   function Hash (Item : Asis.Element) return Ada.Containers.Hash_Type;

   package Asis_Element_Maps is
     new Ada.Containers.Hashed_Maps
          (Asis.Element, Positive, Hash, Asis.Elements.Is_Equal);

   Qt_Types : constant array (Positive range <>) of Qt_Type_Info
     := ((Amoc.Model.Primitive,
          +"bool",                   +"Standard",          +"Boolean"),
         (Amoc.Model.Primitive,
          +"double",                 +"Qt4",               +"Q_Double"),
         (Amoc.Model.Primitive,
          +"float",                  +"Qt4",               +"Q_Float"),
         (Amoc.Model.Primitive,
          +"int",                    +"Qt4",               +"Q_Integer"),
         (Amoc.Model.Primitive,
          +"qint64",                 +"Qt4",               +"Q_Integer_64"),
         (Amoc.Model.Primitive,
          +"qreal",                  +"Qt4",               +"Q_Real"),
         (Amoc.Model.Enumeration,
          +"Qt::DockWidgetArea",     +"Qt4",
          +"Dock_Widget_Area"),
         (Amoc.Model.Value_Class,
          +"QChar",                  +"Qt4.Chars",         +"Q_Char"),
         (Amoc.Model.Value_Class,
          +"QModelIndex",            +"Qt4.Model_Indices", +"Q_Model_Index"),
         (Amoc.Model.Enumeration,
          +"QProcess::ExitStatus",   +"Qt4.Processes",
          +"Process_Exit_Status"),
         (Amoc.Model.Enumeration,
          +"QProcess::ProcessError", +"Qt4.Processes",
          +"Process_Process_Error"),
         (Amoc.Model.Enumeration,
          +"QProcess::ProcessState", +"Qt4.Processes",
          +"Process_Process_State"),
         (Amoc.Model.Value_Class,
          +"QString",                +"Qt4.Strings",       +"Q_String"),
         (Amoc.Model.Value_Class,
          +"QVariant",               +"Qt4.Variants",      +"Q_Variant"),
         (Amoc.Model.QObject_Subclass,
          +"QWidget",                +"Qt4.Widgets",       +"Q_Widget"));

   Type_Map                      : Asis_Element_Maps.Map;
   Q_Object_Director_Declaration : Asis.Element;

   ----------
   -- Hash --
   ----------

   function Hash (Item : Asis.Element) return Ada.Containers.Hash_Type is
      use type Asis.ASIS_Integer;

   begin
      return Ada.Containers.Hash_Type (abs Asis.Elements.Hash (Item));
   end Hash;

   -------------------
   -- Is_Child_Unit --
   -------------------

   function Is_Child_Unit
    (Unit   : Asis.Compilation_Unit;
     Parent : Asis.Compilation_Unit)
       return Boolean
   is
      Aux : Asis.Compilation_Unit
        := Asis.Compilation_Units.Corresponding_Parent_Declaration (Unit);

   begin
      while not Asis.Compilation_Units.Is_Nil (Aux) loop
         if Asis.Compilation_Units.Is_Equal (Aux, Parent) then
            return True;
         end if;

         Aux := Asis.Compilation_Units.Corresponding_Parent_Declaration (Aux);
      end loop;

      return False;
   end Is_Child_Unit;

   ----------------
   -- Is_Derived --
   ----------------

   function Is_Derived
    (Declaration      : Asis.Element;
     Type_Declaration : Asis.Element;
     Only_Public      : Boolean)
       return Boolean
   is
      function Is_Derived_From_Ancestor
       (Type_Definition : Asis.Element)
          return Boolean;
      --  Returns True if type derived from ancestor.

      function Is_Derived_From_Progenitors
       (Type_Definition : Asis.Element)
          return Boolean;
      --  Returns True if type derived from progenitors.

      ------------------------------
      -- Is_Derived_From_Ancestor --
      ------------------------------

      function Is_Derived_From_Ancestor
       (Type_Definition : Asis.Element)
          return Boolean
      is
         Aux : Asis.Element;

      begin
         case Asis.Elements.Definition_Kind (Type_Definition) is
            when Asis.A_Private_Extension_Definition =>
               Aux :=
                 Asis.Definitions.Subtype_Mark
                  (Asis.Definitions.Ancestor_Subtype_Indication
                    (Type_Definition));

            when Asis.A_Type_Definition =>
               case Asis.Elements.Type_Kind (Type_Definition) is
                  when Asis.A_Derived_Record_Extension_Definition =>
                     Aux :=
                       Asis.Definitions.Subtype_Mark
                        (Asis.Definitions.Parent_Subtype_Indication
                          (Type_Definition));

                  when others =>
                     raise Program_Error;
               end case;

            when others =>
               raise Program_Error;
         end case;

         case Asis.Elements.Expression_Kind (Aux) is
            when Asis.A_Selected_Component =>
               Aux := Asis.Expressions.Selector (Aux);

            when Asis.An_Identifier =>
               null;

            when others =>
               raise Program_Error;
         end case;

         Aux := Asis.Expressions.Corresponding_Name_Declaration (Aux);

         --  If parent type declared in the parent unit, when use full type
         --  declaration for recursive Is_Derived query.

         if not Only_Public
           and then Is_Child_Unit
                     (Asis.Elements.Enclosing_Compilation_Unit (Declaration),
                      Asis.Elements.Enclosing_Compilation_Unit (Aux))
         then
            return
              Asis.Elements.Is_Equal (Aux, Type_Declaration)
                or else Is_Derived
                         (Asis.Declarations.Corresponding_Type_Declaration
                           (Aux),
                          Type_Declaration,
                          Only_Public);

         else
            return
              Asis.Elements.Is_Equal (Aux, Type_Declaration)
                or else Is_Derived (Aux, Type_Declaration, Only_Public);
         end if;
      end Is_Derived_From_Ancestor;

      ---------------------------------
      -- Is_Derived_From_Progenitors --
      ---------------------------------

      function Is_Derived_From_Progenitors
       (Type_Definition : Asis.Element)
          return Boolean
      is
         Progenitors : constant Asis.Element_List
           := Amoc.Asis_Implementation.Progenitor_List (Type_Definition);
         Aux         : Asis.Element;

      begin
         for J in Progenitors'Range loop
            Aux := Progenitors (J);

            case Asis.Elements.Expression_Kind (Aux) is
               when Asis.A_Selected_Component =>
                  Aux := Asis.Expressions.Selector (Aux);

               when Asis.An_Identifier =>
                  null;

               when others =>
                  raise Program_Error;
            end case;

            Aux := Asis.Expressions.Corresponding_Name_Declaration (Aux);

            if Asis.Elements.Is_Equal (Aux, Type_Declaration)
              or else Is_Derived (Aux, Type_Declaration, Only_Public)
            then
               return True;
            end if;
         end loop;

         return False;
      end Is_Derived_From_Progenitors;

      Type_Definition : Asis.Element;

   begin
      case Asis.Elements.Declaration_Kind (Declaration) is
         when Asis.A_Private_Extension_Declaration =>
            return
              Is_Derived_From_Ancestor
               (Asis.Declarations.Type_Declaration_View (Declaration));

         when Asis.An_Ordinary_Type_Declaration =>
            Type_Definition :=
              Asis.Declarations.Type_Declaration_View (Declaration);

            case Asis.Elements.Type_Kind (Type_Definition) is
               when Asis.An_Interface_Type_Definition =>
                  return Is_Derived_From_Progenitors (Type_Definition);

               when Asis.A_Derived_Record_Extension_Definition =>
                  return
                    Is_Derived_From_Ancestor (Type_Definition)
                      or else Is_Derived_From_Progenitors (Type_Definition);

               when others =>
                  raise Program_Error;
            end case;

         when others =>
            raise Program_Error;
      end case;
   end Is_Derived;

   ---------------------------------------
   -- Is_Derived_From_Q_Object_Director --
   ---------------------------------------

   function Is_Derived_From_Q_Object_Director
    (Declaration : Asis.Element;
     Only_Public : Boolean)
       return Boolean
   is
   begin
      return
        Is_Derived (Declaration, Q_Object_Director_Declaration, Only_Public);
   end Is_Derived_From_Q_Object_Director;

   ---------------------
   -- Process_Qt_Type --
   ---------------------

   procedure Process_Qt_Type (Info : Qt_Type_Info) is
      use type Asis.Expression_Kinds;
      use type Asis.Unit_Kinds;

      Unit : Asis.Compilation_Unit
        := Asis.Compilation_Units.Library_Unit_Declaration
            (-Info.Ada_Package, Context);

   begin
      if Asis.Compilation_Units.Unit_Kind (Unit)
           = Asis.A_Package_Renaming
      then
         declare
            Renamed : Asis.Element
              := Asis.Declarations.Corresponding_Base_Entity
                  (Asis.Elements.Unit_Declaration (Unit));

         begin
            if Asis.Elements.Expression_Kind (Renamed)
                 = Asis.A_Selected_Component
            then
               Renamed := Asis.Expressions.Selector (Renamed);
            end if;

            Unit :=
              Asis.Elements.Enclosing_Compilation_Unit
               (Asis.Expressions.Corresponding_Name_Definition (Renamed));
         end;
      end if;

      if not Asis.Compilation_Units.Is_Nil (Unit) then
         declare
            use type Asis.Declaration_Kinds;
            use type Asis.Element_Kinds;

            Decl  : constant Asis.Element
              := Asis.Elements.Unit_Declaration (Unit);
            Decls : constant Asis.Element_List
              := Asis.Declarations.Visible_Part_Declarative_Items (Decl);

         begin
            for K in Decls'Range loop
               --  Gela's Asis.Declarations.Names query returns empty
               --  ASIS.Element_List for implicit declarations like
               --  root-integer. Thus, it is important to skip such
               --  declarations.

               if not Asis.Elements.Is_Part_Of_Implicit (Decls (K))
                 and then (Asis.Elements.Declaration_Kind (Decls (K))
                             = Asis.A_Private_Type_Declaration
                   or else Asis.Elements.Declaration_Kind (Decls (K))
                             = Asis.An_Ordinary_Type_Declaration)
               then
                  declare
                     use type Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

                     Names : constant Asis.Element_List
                       := Asis.Declarations.Names (Decls (K));
                     Name  : constant Asis.Element := Names (Names'First);
                     Image : constant Wide_String
                       := Asis.Declarations.Defining_Name_Image (Name);

                  begin
                     if Image = Info.Ada_Name then
                        case Info.Kind is
                           when Amoc.Model.Primitive =>
                              Amoc.Model.Known_Types.Append
                               ((Kind           => Info.Kind,
                                 Qt_Name        => Info.Qt_Name,
                                 Ada_Name       =>
                                   Info.Ada_Package & '.' & Info.Ada_Name,
                                 Create_Wrapper => +"",
                                 To_Pointer     => +"",
                                 To_Address     => +"",
                                 Class_View     => +"",
                                 Utils_Package  => +""));

                           when Amoc.Model.Enumeration =>
                              Amoc.Model.Known_Types.Append
                               ((Kind           => Info.Kind,
                                 Qt_Name        => Info.Qt_Name,
                                 Ada_Name       =>
                                   Info.Ada_Package & '.' & Info.Ada_Name,
                                 Create_Wrapper => +"",
                                 To_Pointer     => +"",
                                 To_Address     => +"",
                                 Class_View     => +"",
                                 Utils_Package  => +""));

                           when Amoc.Model.Value_Class =>
                              Amoc.Model.Known_Types.Append
                               ((Kind           => Info.Kind,
                                 Qt_Name        => Info.Qt_Name,
                                 Ada_Name       =>
                                   Info.Ada_Package & '.' & Info.Ada_Name,
                                 Create_Wrapper =>
                                   Info.Ada_Package
                                     & ".Internals.Create_Wrapper",
                                 To_Pointer     =>
                                   Info.Ada_Package
                                     & ".Internals.To_Pointer",
                                 To_Address     =>
                                   Info.Ada_Package
                                     & ".Internals.To_Address",
                                 Class_View     =>
                                   Info.Ada_Package
                                     & ".Internals."
                                     & Info.Qt_Name
                                     & "_View",
                                 Utils_Package  => +""));

                           when Amoc.Model.QObject_Subclass =>
                              Amoc.Model.Known_Types.Append
                               ((Kind           => Info.Kind,
                                 Qt_Name        => Info.Qt_Name,
                                 Ada_Name       =>
                                   Info.Ada_Package & '.' & Info.Ada_Name,
                                 Create_Wrapper => +"",
                                 To_Pointer     =>
                                   +"Qt_Ada.Internals.Objects.Widget_Utilities"
                                     & ".To_Pointer",
                                 To_Address     =>
                                   +"Qt_Ada.Internals.Objects.Widget_Utilities"
                                     & ".To_Address",
                                 Class_View     => +"",
                                 Utils_Package  =>
                                   +("Qt_Ada.Internals.Objects"
                                       & ".Widget_Utilities")));
                        end case;

                        Type_Map.Insert
                         (Decls (K), Positive (Amoc.Model.Known_Types.Length));
                     end if;
                  end;
               end if;
            end loop;
         end;
      end if;
   end Process_Qt_Type;

   ----------------------
   -- Process_Qt_Types --
   ----------------------

   procedure Process_Qt_Types is

      function Search_For_Type_Declaration
       (Unit_Name : Wide_String;
        Type_Name : Wide_String)
          return Asis.Element;
      --  Returns

      ---------------------------------
      -- Search_For_Type_Declaration --
      ---------------------------------

      function Search_For_Type_Declaration
       (Unit_Name : Wide_String;
        Type_Name : Wide_String)
          return Asis.Element
      is
         Unit : constant Asis.Compilation_Unit
           := Asis.Compilation_Units.Library_Unit_Declaration
               (Unit_Name, Context);

      begin
         if not Asis.Compilation_Units.Is_Nil (Unit) then
            declare
               Decls : constant Asis.Element_List
                 := Asis.Declarations.Visible_Part_Declarative_Items
                     (Asis.Elements.Unit_Declaration (Unit));

            begin
               for J in Decls'Range loop
                  declare
                     Names : constant Asis.Element_List
                       := Asis.Declarations.Names (Decls (J));
                     Image : constant Wide_String
                       := Asis.Declarations.Defining_Name_Image
                           (Names (Names'First));

                  begin
                     if Image = Type_Name then
                        return Decls (J);
                     end if;
                  end;
               end loop;
            end;
         end if;

         return Asis.Nil_Element;
      end Search_For_Type_Declaration;

   begin
      --  Looking for Q_Object_Director type declaration.

      Q_Object_Director_Declaration :=
        Search_For_Type_Declaration
         ("Qt4.Objects.Directors", "Q_Object_Director");

      if Asis.Elements.Is_Nil (Q_Object_Director_Declaration) then
         --  If Qt4.Objects.Q_Object_Director type declaration is not found
         --  than there is no derived types, so nothing to do at all.

         Amoc.Diagnoses.Output
          (Asis.Compilation_Units.Text_Name (Primary_Unit),
           1,
           1,
           Amoc.Diagnoses.Unable_To_Find_Type_Derived_From_Q_Object_Director);

         raise Amoc.Amoc_Error;
      end if;

      --  Process others known Qt types

      for J in Qt_Types'Range loop
         Process_Qt_Type (Qt_Types (J));
      end loop;
   end Process_Qt_Types;

   -------------
   -- Resolve --
   -------------

   function Resolve (Declaration : Asis.Element) return Natural is
      Position : constant Asis_Element_Maps.Cursor
        := Type_Map.Find (Declaration);

   begin
      if Asis_Element_Maps.Has_Element (Position) then
         return Asis_Element_Maps.Element (Position);

      else
         return 0;
      end if;
   end Resolve;

end Amoc.Analizer.Types;
