------------------------------------------------------------------------------
--                                                                          --
--                An Ada2005 binding to the Qt C++ framework                --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2008, 2009 Vadim Godunko <vgodunko@gmail.com>                --
--                                                                          --
-- QtAda is free software;  you can  redistribute it and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. QtAda is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License distributed with QtAda;  see file COPYING.  If not, write --
-- to  the  Free Software Foundation,  51  Franklin  Street,  Fifth  Floor, --
-- Boston, MA 02110-1301, USA.                                              --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 3338 $ $Date: 2009-05-04 12:54:00 +0400 (Mon, 04 May 2009) $
------------------------------------------------------------------------------
with Ada.Characters.Conversions;
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Containers.Indefinite_Hashed_Sets;
with Ada.Strings.Unbounded;
with Ada.Strings.Wide_Fixed;
with Ada.Strings.Wide_Hash;
with Ada.Strings.Wide_Unbounded;

with Asis.Compilation_Units;
with Asis.Declarations;
with Asis.Definitions;
with Asis.Elements;
with Asis.Expressions;

with Amoc.Analizer.Dependencies;
with Amoc.Analizer.Types;
with Amoc.Asis_Implementation;
with Amoc.Diagnoses;
with Amoc.Model;
with Amoc.Parameters;
with Amoc.String_Utilities;

package body Amoc.Analizer is

   use Ada.Strings.Unbounded;
   use Amoc.Model;
   use Asis;
   use Asis.Declarations;
   use Asis.Definitions;
   use Asis.Elements;
   use Asis.Expressions;

   procedure Prepare;
   --  Производит подготовительные действия и формирует информацию, необходимую
   --  для проведения анализа модуля.

   procedure Analyze_Pragma_Q_Object (Pragma_Declaration : Asis.Element);
   --  Analizes pragma Q_Object declaration.

   procedure Analyze_Method_Pragma
    (Pragma_Element : Asis.Element;
     Kind           : Amoc.Model.Method_Kinds);
   --  Analizes pragmas Q_Signal/Q_Slot/Q_Invokable.

   procedure Analyze_Pragma_Q_Property (Pragma_Element : Asis.Element);
   --  Производит анализ директивы компилятора Q_Property. Последний элемент
   --  списка является собственно объявлением директивы компилятора.

   procedure Analyze_Pragma_Q_Export_Plugin (Pragma_Element : Asis.Element);
   --  Производит анализ директивы компилятора Q_Export_Plugin.

   procedure Analyze_Pragma_Q_Declare_Interface
    (Iface : Asis.Element; Decls : Asis.Element_List);
   --  Производит анализ директивы компилятора Q_Declare_Interface.
   --  Последний элемент списка является собственно
   --  объявлением директивы компилятора.

   procedure Analyze_Interfaces;
   --  Производит анализ интерфейсов, которые реализует данный класс.
   --  При анализе интерфейсов производится проверка на наличие прагмы
   --  Q_DECLARE_INTERFACE.

   function Analyze_Parameter_Profile (Subprogram : Asis.Element)
     return Amoc.Model.Parameter_Vectors.Vector;
   --  Analyze subprogram's profile parameters and returns extracted
   --  information.

   function Extended_Name_Image (Element : Asis.Element) return Wide_String;
   --  Returns Name_Image for An_Identifier and Extended_Name_Image (Prefix
   --  (Element)) & '.' & Name_Image (Selector (Element)) for
   --  A_Selected_Component.

   procedure Analyze_Type_Declarations
    (Visible_Declarations : Asis.Element_List);
   --  Analyses type declarations in the visible part of the package
   --  declaration and type completion in the private part of the package
   --  declaration for detect type declaration which declares tagged type
   --  for user defined class. Also, determine parent tagged type and
   --  is Static_Meta_Object object declared in the same package as parent
   --  tagged type or not.

   procedure Analyze_Declaration
    (Declaration : Asis.Element;
     Visibility  : Amoc.Model.Visibility_Kinds);
   --  Analyze declaration. If declaration is a subprogram declaration, then
   --  construct Known_Subprograms and Overloaded_Subprograms.

   procedure Analyze_Pragma (Pragma_Element : Asis.Element);
   --  Analyze pragma declaration. For Amoc pragmas call corresponding
   --  Analyze_Pragma_<Name> subprogram.

   function Compute_Signature
    (Subprogram_Name   : Wide_String;
     Parameters        : Amoc.Model.Parameter_Vectors.Vector;
     Requested_Name    : Wide_String;
     Requested_Profile : Wide_String)
       return Ada.Strings.Unbounded.Unbounded_String;
   --  Returns complete signature of the signal/slot/method.

   function To_Cpp_Method_Name (Name : Wide_String) return Wide_String;
   --  Converts name from Ada conventions to C++ conventions.

   function "+"
    (Item : Wide_String)
       return Ada.Strings.Wide_Unbounded.Unbounded_Wide_String
         renames Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String;

   package Wide_String_Sets is
     new Ada.Containers.Indefinite_Hashed_Sets
          (Wide_String, Ada.Strings.Wide_Hash, "=");

   type Subprogram_Info is record
      Element    : Asis.Element;
      Visibility : Amoc.Model.Visibility_Kinds;
   end record;

   package Wide_String_Subprogram_Info_Maps is
     new Ada.Containers.Indefinite_Hashed_Maps
          (Wide_String,
           Subprogram_Info,
           Ada.Strings.Wide_Hash,
           "=");

   Tagged_Type_Uppercase_Name :
     Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
   --  Имя тэгового типа (в верхнем регистре).

   Tagged_Type_Declaration     : Asis.Element;
   --  Asis.Element of the tagged type declaration.
   Pragma_Q_Object_Declaration : Asis.Element;
   --  Asis.Element of the pragma Q_Object declaration.

   Known_Subprograms      : Wide_String_Subprogram_Info_Maps.Map;
   Overloaded_Subprograms : Wide_String_Sets.Set;

   -------------
   -- Analyze --
   -------------

   procedure Analyze is
      use type Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

   begin
      Prepare;

      declare
         Unit_Declaration     : constant Asis.Element
           := Asis.Elements.Unit_Declaration (Primary_Unit);
         Visible_Declarations : constant Asis.Element_List
           := Asis.Declarations.Visible_Part_Declarative_Items
               (Unit_Declaration, True);
         Private_Declarations : constant Asis.Element_List
           := Asis.Declarations.Private_Part_Declarative_Items
               (Unit_Declaration, True);

      begin
         Analyze_Type_Declarations (Visible_Declarations);

         --  Анализ наследуемых интерфейсов у которых определена
         --  pragma Q_DECLARE_INTERFACE

         Analyze_Interfaces;

         --  Основной проход: поиск и анализ директив компилятора.

         for J in Visible_Declarations'Range loop
            case Asis.Elements.Element_Kind (Visible_Declarations (J)) is
               when Asis.A_Declaration =>
                  Analyze_Declaration
                   (Visible_Declarations (J), Amoc.Model.A_Public);

               when Asis.A_Pragma =>
                  Analyze_Pragma (Visible_Declarations (J));

               when others =>
                  null;
            end case;
         end loop;

         for J in Private_Declarations'Range loop
            case Asis.Elements.Element_Kind (Private_Declarations (J)) is
               when Asis.A_Declaration =>
                  Analyze_Declaration
                   (Private_Declarations (J), Amoc.Model.A_Protected);

               when Asis.A_Pragma =>
                  Analyze_Pragma (Private_Declarations (J));

               when others =>
                  null;
            end case;
         end loop;
      end;

      --  Compute additional information for code generator.

      declare
         Names : constant Asis.Element_List
           := Asis.Declarations.Names (Tagged_Type_Declaration);

      begin
         Amoc.Model.Tagged_Type_Name :=
           +Asis.Declarations.Defining_Name_Image (Names (Names'First));
      end;

      --  If custom class name not defined, then use default.

      if Amoc.Model.Custom_Class_Name = "" then
         Amoc.Model.Custom_Class_Name := Amoc.Model.Tagged_Type_Name;
      end if;

      --  Analyze dependencies

      if Amoc.Parameters.Output_Dependencies then
         Amoc.Analizer.Dependencies.Analyze;
      end if;

      if Amoc.Parameters.Verbose then
         Amoc.Model.Output_Summary;
      end if;
   end Analyze;

   -------------------------
   -- Analyze_Declaration --
   -------------------------

   procedure Analyze_Declaration
    (Declaration : Asis.Element;
     Visibility  : Amoc.Model.Visibility_Kinds)
   is
   begin
      case Declaration_Kind (Declaration) is
         when A_Procedure_Declaration
           | A_Function_Declaration
         =>
            declare
               Names     : constant Asis.Element_List
                 := Asis.Declarations.Names (Declaration);
               Name      : constant Wide_String
                 := Amoc.String_Utilities.To_Upper
                     (Asis.Declarations.Defining_Name_Image
                       (Names (Names'First)));
               Profile   : constant Asis.Element_List
                 := Asis.Declarations.Parameter_Profile (Declaration);
               Aux       : Asis.Element;
               Candidate : Boolean := False;

            begin
               if Profile'Length /= 0 then
                  --  Check type of the first parameter.

                  Aux :=
                    Amoc.Asis_Implementation.Object_Declaration_Subtype
                     (Profile (Profile'First));

                  case Asis.Elements.Element_Kind (Aux) is
                     when Asis.A_Definition =>
                        case Asis.Elements.Definition_Kind (Aux) is
                           when Asis.An_Access_Definition =>
                              case Asis.Elements.Access_Definition_Kind
                                    (Aux)
                              is
                                 when Asis.An_Anonymous_Access_To_Variable
                                   | Asis.An_Anonymous_Access_To_Constant
                                 =>
                                    Aux :=
                                      Asis.Definitions.
                                        Anonymous_Access_To_Object_Subtype_Mark
                                         (Aux);

                                 when others =>
                                    Aux := Asis.Nil_Element;
                              end case;

                           when Asis.A_Subtype_Indication =>
                              if Amoc.Asis_Implementation.Asis_Gnat then
                                 --  GNAT ASIS does not have implementation of
                                 --  the Object_Declaration_Subtype subprogram,
                                 --  thus control flow can't be here.

                                 raise Program_Error;
                              end if;

                              Aux := Asis.Definitions.Subtype_Mark (Aux);

                           when others =>
                              raise Program_Error;
                        end case;

                     when Asis.An_Expression =>
                        if not Amoc.Asis_Implementation.Asis_Gnat then
                           --  Gela ASIS has implementation of the
                           --  Object_Declaration_Subtype subprogram, thus
                           --  control flow can't be here.

                           raise Program_Error;
                        end if;

                        case Asis.Elements.Expression_Kind (Aux) is
                           when Asis.An_Attribute_Reference =>
                              case Asis.Elements.Attribute_Kind (Aux) is
                                 when Asis.A_Class_Attribute =>
                                    null;

                                 when others =>
                                    raise Program_Error;
                              end case;

                           when Asis.An_Identifier
                             | Asis.A_Selected_Component
                           =>
                              null;

                           when others =>
                              raise Program_Error;
                        end case;

                     when others =>
                        raise Program_Error;
                  end case;

                  if not Asis.Elements.Is_Nil (Aux) then
                     --  Unwind 'Class attribute if any.

                     case Asis.Elements.Expression_Kind (Aux) is
                        when Asis.An_Attribute_Reference =>
                           case Asis.Elements.Attribute_Kind (Aux) is
                              when Asis.A_Class_Attribute =>
                                 Aux := Asis.Expressions.Prefix (Aux);

                              when others =>
                                 raise Program_Error;
                           end case;

                        when Asis.An_Identifier
                          | Asis.A_Selected_Component
                        =>
                           null;

                        when others =>
                           raise Program_Error;
                     end case;

                     --  Unwind prefixes.

                     case Asis.Elements.Expression_Kind (Aux) is
                        when Asis.A_Selected_Component =>
                           Aux := Asis.Expressions.Selector (Aux);

                        when Asis.An_Identifier =>
                           null;

                        when others =>
                           raise Program_Error;
                     end case;

                     Aux :=
                       Asis.Expressions.Corresponding_Name_Declaration (Aux);

                     if Asis.Elements.Is_Equal (Aux, Tagged_Type_Declaration)
                       or else Asis.Elements.Is_Equal
                                (Aux,
                                 Asis.Declarations.
                                   Corresponding_Type_Declaration
                                    (Tagged_Type_Declaration))
                     then
                        Candidate := True;
                     end if;
                  end if;
               end if;

               if Candidate
                 and then not Overloaded_Subprograms.Contains (Name)
               then
                  if Known_Subprograms.Contains (Name) then
                     Overloaded_Subprograms.Insert (Name);
                     Known_Subprograms.Delete (Name);

                  else
                     Known_Subprograms.Insert
                      (Name, (Declaration, Visibility));
                  end if;
               end if;
            end;

         when others =>
            null;
      end case;
   end Analyze_Declaration;

   ------------------------
   -- Analyze_Interfaces --
   ------------------------
   procedure Analyze_Interfaces is
      Interfaces : constant Asis.Expression_List
        := Amoc.Asis_Implementation.Progenitor_List
            (Type_Declaration_View (Tagged_Type_Declaration));
      --  Получаем список интерфейсов, которые реализует данный класс.

      Item : Asis.Element;
      Unit : Asis.Compilation_Unit;

   begin
      for J in Interfaces'Range loop
         if Expression_Kind (Interfaces (J)) = A_Selected_Component then
            Item := Corresponding_Name_Declaration
                     (Selector (Interfaces (J)));

         else
            Item := Corresponding_Name_Declaration (Interfaces (J));
         end if;

         --  Получаем declaration интерфейса.

         Unit := Enclosing_Compilation_Unit (Item);

         declare
            Decl  : constant Asis.Element := Unit_Declaration (Unit);
            Decls : constant Asis.Element_List
              := Asis.Declarations.Visible_Part_Declarative_Items (Decl, True);
            --  Получаем все описания, находящиеся в модуле.

         begin
            for J in Decls'Range loop
               if Element_Kind (Decls (J)) = A_Pragma
                 and then Pragma_Kind (Decls (J)) = An_Unknown_Pragma
               then
                  --  Находим описание прагмы Q_Declare_Interface.

                  declare
                     Name_Image : constant Wide_String
                       := Amoc.String_Utilities.To_Upper
                           (Pragma_Name_Image (Decls (J)));

                  begin
                     if Name_Image = "Q_DECLARE_INTERFACE" then
                        Analyze_Pragma_Q_Declare_Interface
                         (Item, Decls (Decls'First .. J));
                     end if;
                  end;
               end if;
            end loop;
         end;
      end loop;
   end Analyze_Interfaces;

   ---------------------------
   -- Analyze_Method_Pragma --
   ---------------------------

   --  pragma Q_Signal (Subprogram => Subprogram_Identifier,
   --                   Signature  => String_Literal);
   --
   --  pragma Q_Slot (Subprogram => Subprogram_Identifier,
   --                 Signature  => String_Literal);
   --
   --  pragma Q_Invokable (Subprogram => Subprogram_Identifier,
   --                      Signature  => String_Literal);

   procedure Analyze_Method_Pragma
    (Pragma_Element : Asis.Element;
     Kind           : Amoc.Model.Method_Kinds)
   is
      Is_Signal              : constant Boolean := Kind = Amoc.Model.Signal;
      Args                   : constant Asis.Element_List
        := Asis.Elements.Pragma_Argument_Associations (Pragma_Element);
      Subprogram_Identifier  : Asis.Element;
      Signal_Signature       : Asis.Element;
      Subprogram_Declaration : Asis.Element;
      Subprogram_Visibility  : Amoc.Model.Visibility_Kinds;
      Subprogram_Name        :
        Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Method_Name            :
        Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Requested_Name         :
        Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Requested_Profile      :
        Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

   begin
      if Args'Length not in 1 .. 2 then
         case Kind is
            when Amoc.Model.Signal =>
               Amoc.Diagnoses.Output
                (Pragma_Element,
                 Amoc.Diagnoses.Q_Signal_Must_Have_1_Or_2_Parameters);

            when Amoc.Model.Slot =>
               Amoc.Diagnoses.Output
                (Pragma_Element,
                 Amoc.Diagnoses.Q_Slot_Must_Have_1_Or_2_Parameters);

            when Amoc.Model.Method =>
               Amoc.Diagnoses.Output
                (Pragma_Element,
                 Amoc.Diagnoses.Q_Invokable_Must_Have_1_Or_2_Parameters);
         end case;
      end if;

      --  Analyze first parameter - subprogram identifier.

      Subprogram_Identifier :=
        Asis.Expressions.Actual_Parameter (Args (Args'First));

      case Asis.Elements.Expression_Kind (Subprogram_Identifier) is
         when An_Identifier =>
            null;

         when others =>
            case Kind is
               when Amoc.Model.Signal =>
                  Amoc.Diagnoses.Output
                    (Subprogram_Identifier,
                     Amoc.Diagnoses.Q_Signal_1st_Must_Be_Direct_Name);

               when Amoc.Model.Slot =>
                  Amoc.Diagnoses.Output
                    (Subprogram_Identifier,
                     Amoc.Diagnoses.Q_Slot_1st_Must_Be_Direct_Name);

               when Amoc.Model.Method =>
                  Amoc.Diagnoses.Output
                    (Subprogram_Identifier,
                     Amoc.Diagnoses.Q_Invokable_1st_Must_Be_Direct_Name);
            end case;
      end case;

      Subprogram_Name :=
        Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
         (Asis.Expressions.Name_Image (Subprogram_Identifier));

      declare
         Uppercase_Name : constant Wide_String
           := Amoc.String_Utilities.To_Upper
               (Ada.Strings.Wide_Unbounded.To_Wide_String (Subprogram_Name));

      begin
         if Is_Signal then
            if Uppercase_Name'Length <= 5
              or else Uppercase_Name
                       (Uppercase_Name'First .. Uppercase_Name'First + 4)
                /= "EMIT_"
            then
               Amoc.Diagnoses.Output
                (Subprogram_Identifier,
                 Amoc.Diagnoses.Q_Signal_Subprogram_Must_Start_From_Emit);
            end if;

            Method_Name :=
              Ada.Strings.Wide_Unbounded.Unbounded_Slice
               (Subprogram_Name,
                6,
                Ada.Strings.Wide_Unbounded.Length (Subprogram_Name));

         else
            Method_Name := Subprogram_Name;
         end if;

         if Known_Subprograms.Contains (Uppercase_Name) then
            Subprogram_Declaration :=
              Known_Subprograms.Element (Uppercase_Name).Element;
            Subprogram_Visibility :=
              Known_Subprograms.Element (Uppercase_Name).Visibility;

         else
            case Kind is
               when Amoc.Model.Signal =>
                  Amoc.Diagnoses.Output
                   (Subprogram_Identifier,
                    Amoc.Diagnoses.Q_Signal_Subprogram_Not_Found);

               when Amoc.Model.Slot =>
                  Amoc.Diagnoses.Output
                   (Subprogram_Identifier,
                    Amoc.Diagnoses.Q_Slot_Subprogram_Not_Found);

               when Amoc.Model.Method =>
                  Amoc.Diagnoses.Output
                   (Subprogram_Identifier,
                    Amoc.Diagnoses.Q_Invokable_Subprogram_Not_Found);
            end case;
         end if;

         case Asis.Elements.Declaration_Kind (Subprogram_Declaration) is
            when A_Procedure_Declaration =>
               null;

            when others =>
               if Is_Signal then
                  Amoc.Diagnoses.Output
                   (Subprogram_Declaration,
                    Amoc.Diagnoses.Q_Signal_Subprogram_Must_Be_Procedure);
               end if;
         end case;
      end;

      --  Analyze second parameter - method's signature.

      if Args'Length = 2 then
         Signal_Signature :=
           Asis.Expressions.Actual_Parameter (Args (Args'First + 1));

         case Asis.Elements.Expression_Kind (Signal_Signature) is
            when Asis.A_String_Literal =>
               null;

            when others =>
               case Kind is
                  when Amoc.Model.Signal =>
                     Amoc.Diagnoses.Output
                      (Signal_Signature,
                       Amoc.Diagnoses.Q_Signal_2nd_Must_Be_String_Literal);

                  when Amoc.Model.Slot =>
                     Amoc.Diagnoses.Output
                      (Signal_Signature,
                       Amoc.Diagnoses.Q_Slot_2nd_Must_Be_String_Literal);

                  when Amoc.Model.Method =>
                     Amoc.Diagnoses.Output
                      (Signal_Signature,
                       Amoc.Diagnoses.Q_Invokable_2nd_Must_Be_String_Literal);
               end case;
         end case;

         declare
            Aux                    : constant Wide_String
              := Asis.Expressions.Value_Image (Signal_Signature);
            Signal_Signature_Image : constant Wide_String
              := Aux (Aux'First + 1 .. Aux'Last - 1);
            Index                  : constant Natural
              := Ada.Strings.Wide_Fixed.Index (Signal_Signature_Image, "(");

         begin
            if Index = 0 then
               Requested_Name :=
                 Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
                  (Signal_Signature_Image);

            else
               Requested_Name :=
                 Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
                  (Signal_Signature_Image
                    (Signal_Signature_Image'First .. Index - 1));
               Requested_Profile :=
                 Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
                  (Signal_Signature_Image
                    (Index .. Signal_Signature_Image'Last));
            end if;
         end;
      end if;

      declare
         Parameters : constant Amoc.Model.Parameter_Vectors.Vector
           := Analyze_Parameter_Profile (Subprogram_Declaration);

      begin
         case Kind is
            when Amoc.Model.Signal =>
               Amoc.Model.Signals.Append
                ((Subprogram_Name =>
                    Ada.Strings.Unbounded.To_Unbounded_String
                     (Ada.Characters.Conversions.To_String
                       (Ada.Strings.Wide_Unbounded.To_Wide_String
                         (Subprogram_Name))),
                  Signature       =>
                    Compute_Signature
                     (Ada.Strings.Wide_Unbounded.To_Wide_String (Method_Name),
                      Parameters,
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Name),
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Profile)),
                  Parameters      => Parameters,
                  Visibility      => Subprogram_Visibility,
                  Kind            => Amoc.Model.Signal));

            when Amoc.Model.Slot =>
               Amoc.Model.Slots.Append
                ((Subprogram_Name =>
                    Ada.Strings.Unbounded.To_Unbounded_String
                     (Ada.Characters.Conversions.To_String
                       (Ada.Strings.Wide_Unbounded.To_Wide_String
                         (Subprogram_Name))),
                  Signature       =>
                    Compute_Signature
                     (Ada.Strings.Wide_Unbounded.To_Wide_String (Method_Name),
                      Parameters,
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Name),
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Profile)),
                  Parameters      => Parameters,
                  Visibility      => Subprogram_Visibility,
                  Kind            => Amoc.Model.Slot));

            when Amoc.Model.Method =>
               Amoc.Model.Methods.Append
                ((Subprogram_Name =>
                    Ada.Strings.Unbounded.To_Unbounded_String
                     (Ada.Characters.Conversions.To_String
                       (Ada.Strings.Wide_Unbounded.To_Wide_String
                         (Subprogram_Name))),
                  Signature       =>
                    Compute_Signature
                     (Ada.Strings.Wide_Unbounded.To_Wide_String (Method_Name),
                      Parameters,
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Name),
                      Ada.Strings.Wide_Unbounded.To_Wide_String
                       (Requested_Profile)),
                  Parameters      => Parameters,
                  Visibility      => Subprogram_Visibility,
                  Kind            => Amoc.Model.Method));
         end case;
      end;
   end Analyze_Method_Pragma;

   -------------------------------
   -- Analyze_Parameter_Profile --
   -------------------------------

   function Analyze_Parameter_Profile (Subprogram : Asis.Element)
     return Amoc.Model.Parameter_Vectors.Vector
   is
      Parameters : constant Asis.Parameter_Specification_List
        := Asis.Declarations.Parameter_Profile (Subprogram);
      --  Obtain list of the subprogram's parameters declarations.

      Result     : Amoc.Model.Parameter_Vectors.Vector;

   begin
      for J in Parameters'Range loop
         declare
            Names        : constant Asis.Element_List
              := Asis.Declarations.Names (Parameters (J));
            Name         : constant Wide_String
              := Asis.Declarations.Defining_Name_Image (Names (Names'First));
            View         : constant Asis.Element
              := Amoc.Asis_Implementation.Object_Declaration_Subtype
                  (Parameters (J));
            Type_Image   : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
            Type_Mark    : Asis.Element;
            Is_Access    : Boolean;
            Is_Constant  : Boolean;
            Is_Not_Null  : Boolean;
            Is_Classwide : Boolean := False;

         begin
            case Asis.Elements.Access_Definition_Kind (View) is
               when An_Anonymous_Access_To_Variable =>
                  Is_Access   := True;
                  Is_Constant := False;
                  Is_Not_Null :=
                    Amoc.Asis_Implementation.Has_Null_Exclusion (View);

               when An_Anonymous_Access_To_Constant =>
                  Is_Access   := True;
                  Is_Constant := True;
                  Is_Not_Null :=
                    Amoc.Asis_Implementation.Has_Null_Exclusion (View);

               when others =>
                  Is_Access   := False;
                  Is_Constant := True;
                  Is_Not_Null := True;
            end case;

            if Is_Access then
               Type_Mark := Anonymous_Access_To_Object_Subtype_Mark (View);

            else
               if Amoc.Asis_Implementation.Asis_Gnat then
                  Type_Mark := View;

               else
                  Type_Mark := Asis.Definitions.Subtype_Mark (View);
               end if;
            end if;

            if Asis.Elements.Attribute_Kind (Type_Mark)
                    = Asis.A_Class_Attribute
            then
               Is_Classwide := True;
               Type_Mark := Asis.Expressions.Prefix (Type_Mark);
            end if;

            if J = Parameters'First then
               Type_Image :=
                 Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
                  (Extended_Name_Image (Type_Mark));
            end if;

            if Asis.Elements.Expression_Kind (Type_Mark)
                 = Asis.A_Selected_Component
            then
               Type_Mark := Asis.Expressions.Selector (Type_Mark);
            end if;

            Result.Append
             ((Name          =>
                 Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String (Name),
               Type_Image    => Type_Image,
               Type_Id       =>
                 Amoc.Analizer.Types.Resolve
                  (Asis.Expressions.Corresponding_Name_Declaration
                    (Type_Mark)),
               Is_Classwide  => Is_Classwide,
               Is_Access     => Is_Access,
               Is_Constant   => Is_Constant,
               Is_Not_Null   => Is_Not_Null));
         end;
      end loop;

      return Result;
   end Analyze_Parameter_Profile;

   --------------------
   -- Analyze_Pragma --
   --------------------

   procedure Analyze_Pragma (Pragma_Element : Asis.Element) is
   begin
      case Asis.Elements.Pragma_Kind (Pragma_Element) is
         when Asis.An_Unknown_Pragma =>
            declare
               Name_Image : constant Wide_String
                 := Amoc.String_Utilities.To_Upper
                     (Pragma_Name_Image (Pragma_Element));

            begin
               if Name_Image = "Q_SIGNAL" then
                  Analyze_Method_Pragma (Pragma_Element, Amoc.Model.Signal);

               elsif Name_Image = "Q_SLOT" then
                  Analyze_Method_Pragma (Pragma_Element, Amoc.Model.Slot);

               elsif Name_Image = "Q_INVOKABLE" then
                  Analyze_Method_Pragma (Pragma_Element, Amoc.Model.Method);

               elsif Name_Image = "Q_PROPERTY" then
                  Analyze_Pragma_Q_Property (Pragma_Element);

               elsif Name_Image = "Q_EXPORT_PLUGIN" then
                  Analyze_Pragma_Q_Export_Plugin (Pragma_Element);

               elsif Name_Image = "Q_OBJECT" then
                  Analyze_Pragma_Q_Object (Pragma_Element);

               else
                  Amoc.Diagnoses.Output
                   (Pragma_Element, Amoc.Diagnoses.Unknown_Pragma);
               end if;
            end;

         when others =>
            null;
      end case;
   end Analyze_Pragma;

   ----------------------------------------
   -- Analyze_Pragma_Q_Declare_Interface --
   ----------------------------------------
   --  pragma Q_Declare_Interface (Iface      => Interface_Identifier,
   --                              Identifier => Subprogram_Identifier);
   procedure Analyze_Pragma_Q_Declare_Interface
    (Iface : Asis.Element; Decls : Asis.Element_List)
   is
      Args : constant Asis.Element_List
        := Pragma_Argument_Associations (Decls (Decls'Last));

   begin
      if Args'Length /= 2 then
         Amoc.Diagnoses.Raise_Exception
          (Decls (Decls'Last),
           "Q_Declare_Interface pragma must have 2 arguments.");
      end if;

      declare
         Class_Element      : constant Asis.Element
           := Actual_Parameter (Args (Args'First));
         Identifier_Element : constant Asis.Element
           := Actual_Parameter (Args (Args'Last));
         Name               : constant Element_List := Names (Iface);

      begin
         --  Проверки типов аргументов.

         if Expression_Kind (Class_Element) /= An_Identifier then
            Amoc.Diagnoses.Raise_Exception
             (Class_Element,
              "First argument of Q_Declare_Interface pragma "
                & "must be a interface type's direct identifier.");
         end if;

         if Expression_Kind (Identifier_Element) /= A_String_Literal then
            Amoc.Diagnoses.Raise_Exception
             (Identifier_Element,
              "Second argument of Q_Declare_Interface pragma "
                & "must be a string literal.");
         end if;

         declare
            Identifier_Name : constant String
              := Ada.Characters.Conversions.To_String
                  (Value_Image (Identifier_Element));
            Image           : constant Wide_String
              := Amoc.String_Utilities.To_Upper
                  (Asis.Declarations.Defining_Name_Image (Name (Name'First)));

         begin
            --  Поиск интерфейса, имеющего имя, указанное в директиве
            --  компилятора.

            if Amoc.String_Utilities.To_Upper
                (Asis.Expressions.Name_Image (Class_Element)) /= Image
            then
               Amoc.Diagnoses.Raise_Exception
                (Class_Element,
                 "First argument of Q_Declare_Interface pragma "
                   & "not an interface type.");
            end if;

            Declared_Interfaces.Append
             ((Name        => To_Unbounded_String
                               (Ada.Characters.Conversions.To_String
                                 (Name_Image (Class_Element))),
               Identifier  => To_Unbounded_String (Identifier_Name)));
         end;
      end;
   end Analyze_Pragma_Q_Declare_Interface;

   ------------------------------------
   -- Analyze_Pragma_Q_Export_Plugin --
   ------------------------------------
   --  В Qt4 имеется pragma Q_EXPORT_PLUGIN2 ( PluginName, ClassName )
   --  В Ada будет использована Q_Export_Plugin
   --                            (Class       => Class,
   --                             Constructor => Constructor)
   --  XXX Сделать проверку существования и формата функции создания
   --  экземпляра плагина (2-ой параметр прагмы).
   procedure Analyze_Pragma_Q_Export_Plugin (Pragma_Element : Asis.Element) is
      use type Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

      Args           : constant Asis.Element_List
        := Pragma_Argument_Associations (Pragma_Element);
      First_Arg      : constant Asis.Element
        := Actual_Parameter (Args (Args'First));
      Last_Arg       : constant Asis.Element
        := Actual_Parameter (Args (Args'Last));

   begin
      if Args'Length /= 2 then
         Amoc.Diagnoses.Raise_Exception
          (Pragma_Element,
           "pragma Q_Export_Plugin expects two arguments");
      end if;

      if Expression_Kind (First_Arg) /= An_Identifier then
         Amoc.Diagnoses.Raise_Exception
          (First_Arg,
           "Required a tagged type name as the first argument "
             & "of pragma Q_Export_Plugin");
      end if;

      if Expression_Kind (Last_Arg) = A_Selected_Component then
         Plugin.Constructor :=
           Ada.Strings.Unbounded.To_Unbounded_String
            (Ada.Characters.Conversions.To_String
              (Name_Image (Selector (Last_Arg))));

         declare
            Tmp_Element : Asis.Element := Prefix (Last_Arg);

         begin
            while Expression_Kind (Tmp_Element) = A_Selected_Component loop
               Plugin.Constructor :=
                 Ada.Strings.Unbounded.To_Unbounded_String
                  (Ada.Characters.Conversions.To_String
                    (Name_Image (Selector (Tmp_Element))))
                   & '.'
                   & Plugin.Constructor;
               Tmp_Element := Prefix (Tmp_Element);
            end loop;

            Plugin.Constructor :=
              Ada.Strings.Unbounded.To_Unbounded_String
               (Ada.Characters.Conversions.To_String
                 (Name_Image (Tmp_Element)))
                & '.'
                & Plugin.Constructor;
         end;

         --  XXX  Реализовать проверку существования указанной функции
         --  создания экземпляра плагина.

      else
         if Expression_Kind (Last_Arg) /= An_Identifier then
            Amoc.Diagnoses.Raise_Exception
             (Last_Arg,
              "Required a name of function with to plugin "
                & "instance access return value type as the "
                & "second argument of pragma Q_Export_Plugin");
         end if;
      end if;

      if Amoc.String_Utilities.To_Upper
          (Asis.Expressions.Name_Image (First_Arg))
          /= Tagged_Type_Uppercase_Name
      then
         if not Plugin.Is_Declared then
            Plugin.Name :=
              Ada.Strings.Unbounded.To_Unbounded_String
               (Ada.Characters.Conversions.To_String
                 (Name_Image (First_Arg)));
            Plugin.Is_Declared := True;

         else
            Amoc.Diagnoses.Raise_Exception
             (Pragma_Element,
              "There is only one plugin declaration with the "
                & "pragma Q_Export_Plugin is allowable");
         end if;
      end if;

      if not Plugin.Is_Declared then
         Amoc.Diagnoses.Raise_Exception
          (First_Arg,
           "The first pragma Q_Export_Plugin argument "
             & "requires private extended type.");
      end if;
   end Analyze_Pragma_Q_Export_Plugin;

   -----------------------------
   -- Analyze_Pragma_Q_Object --
   -----------------------------

   procedure Analyze_Pragma_Q_Object (Pragma_Declaration : Asis.Element) is

      function Is_Same_Name
       (Identifier  : Asis.Element;
        Declaration : Asis.Element)
          return Boolean;
      --  Returns True if Identifier can identify defining name of the
      --  declaration.

      ------------------
      -- Is_Same_Name --
      ------------------

      function Is_Same_Name
       (Identifier  : Asis.Element;
        Declaration : Asis.Element)
          return Boolean
      is
         Identifier_Image : constant Wide_String
           := Amoc.String_Utilities.To_Upper
               (Asis.Expressions.Name_Image (Identifier));
         Names            : constant Asis.Element_List
           := Asis.Declarations.Names (Declaration);

      begin
         for J in Names'Range loop
            if Identifier_Image
                 = Amoc.String_Utilities.To_Upper
                    (Asis.Declarations.Defining_Name_Image (Names (J)))
            then
               return True;
            end if;
         end loop;

         return False;
      end Is_Same_Name;

      Args       : constant Asis.Element_List
        := Asis.Elements.Pragma_Argument_Associations (Pragma_Declaration);
      Type_Name  : Asis.Element;
      Class_Name : Asis.Element;

   begin
      --  Check for pragma Q_Object is declared only once.

      if not Asis.Elements.Is_Nil (Pragma_Q_Object_Declaration) then
         Amoc.Diagnoses.Output
          (Pragma_Declaration,
           Amoc.Diagnoses.Duplicate_Pragma_Q_Object_Declaration);

         raise Amoc.Amoc_Error;

      else
         Pragma_Q_Object_Declaration := Pragma_Declaration;
      end if;

      --  Check number of parameters.

      if Args'Length /= 2 then
         Amoc.Diagnoses.Output
          (Pragma_Declaration,
           Amoc.Diagnoses.Pragma_Q_Object_Must_Have_2_Arguments);

         raise Amoc.Amoc_Error;
      end if;

      --  Check first prameter: it must be direct identifier and must identify
      --  processed tagged type.

      Type_Name := Asis.Expressions.Actual_Parameter (Args (Args'First));

      if Asis.Elements.Expression_Kind (Type_Name) /= Asis.An_Identifier then
         Amoc.Diagnoses.Output
          (Type_Name,
           Amoc.Diagnoses.First_Argument_Pragma_Q_Object_Must_Be_Identifier);

         raise Amoc.Amoc_Error;
      end if;

      if not Is_Same_Name (Type_Name, Tagged_Type_Declaration) then
         Amoc.Diagnoses.Output
          (Type_Name,
           Amoc.Diagnoses.First_Argument_Pragma_Q_Object_Must_Be_Type_Name);

         raise Amoc.Amoc_Error;
      end if;

      --  Check second parameter: it must be string literal and valid C++
      --  identifier.

      Class_Name := Asis.Expressions.Actual_Parameter (Args (Args'Last));

      if Asis.Elements.Expression_Kind (Class_Name)
           /= Asis.A_String_Literal
      then
         Amoc.Diagnoses.Output
          (Class_Name,
           Amoc.Diagnoses.Second_Argument_Pragma_Q_Object_Must_Be_String);

         raise Amoc.Amoc_Error;
      end if;

      --  XXX Check for be valid C++ identifier is not implemented for now.

      declare
         Class_Name_Image : constant Wide_String
           := Asis.Expressions.Value_Image (Class_Name);

      begin
         if Class_Name_Image'Length = 2 then
            Amoc.Diagnoses.Output
             (Class_Name,
              Amoc.Diagnoses.Second_Argument_Pragma_Q_Object_Must_Be_Valid);

            raise Amoc.Amoc_Error;
         end if;

         Custom_Class_Name :=
           +(Class_Name_Image
              (Class_Name_Image'First + 1 .. Class_Name_Image'Last - 1));
      end;
   end Analyze_Pragma_Q_Object;

   -------------------------------
   -- Analyze_Pragma_Q_Property --
   -------------------------------
   --  pragma Q_Property (My_Widget_Impl,
   --                     myState,
   --                     READ  => getState,
   --                     Write => setState);
   --      Атрибут READ обязательный.
   --      Необходимо реализовать обработку объявления свойства Q_Property
   --      без явного объявления READ, Write.

   procedure Analyze_Pragma_Q_Property (Pragma_Element : Asis.Element) is
      use type Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

      Args : constant Asis.Element_List
        := Asis.Elements.Pragma_Argument_Associations (Pragma_Element);

      function Find_Property_Method (Element : Asis.Element)
         return Asis.Element;

      --------------------------
      -- Find_Property_Method --
      --------------------------

      function Find_Property_Method (Element : Asis.Element)
         return Asis.Element
      is
         Subprogram_Name : constant Wide_String
           := Amoc.String_Utilities.To_Upper
               (Asis.Expressions.Name_Image (Element));

         Result : Asis.Element := Asis.Nil_Element;

      begin
         if Known_Subprograms.Contains (Subprogram_Name) then
            Result := Known_Subprograms.Element (Subprogram_Name).Element;
         end if;

         if Is_Nil (Result) then
            Amoc.Diagnoses.Raise_Exception
             (Element,
              "Subprogram '"
                & Name_Image (Element)
                & "' specified in the Q_Property pragma not found.");
         end if;

         return Result;
      end Find_Property_Method;

   begin
      if Args'Length < 3 then
         Amoc.Diagnoses.Raise_Exception
          (Pragma_Element,
           "Q_Property pragma must have at least three arguments.");
      end if;

      declare
         Class_Element      : constant Asis.Element
           := Actual_Parameter (Args (Args'First));
         Property_Element   : constant Asis.Element
           := Actual_Parameter (Args (Args'First + 1));
         Result             : Property_Info;

         Element_Spec : Asis.Element := Asis.Nil_Element;
         Element_Name : Asis.Element := Asis.Nil_Element;

      begin
         if Amoc.String_Utilities.To_Upper
             (Asis.Expressions.Name_Image (Class_Element))
             /= Tagged_Type_Uppercase_Name
         then
            Amoc.Diagnoses.Raise_Exception
             (Class_Element,
              "First argument of Q_Property pragma"
                & " must be tagged type's direct name.");
         end if;

         if Expression_Kind (Property_Element) /= An_Identifier then
            Amoc.Diagnoses.Raise_Exception
             (Property_Element,
              "Second argument of Q_Property pragma"
                & " must be an identifier.");
         end if;

         for J in Args'First + 2 .. Args'Last loop

            Element_Name := Actual_Parameter (Args (J));
            Element_Spec := Formal_Parameter (Args (J));

            declare
               Element_Image : constant Wide_String
                 := Amoc.String_Utilities.To_Upper
                     (Asis.Expressions.Name_Image (Element_Spec));

               Element_Name_Image : constant Wide_String
                 := Amoc.String_Utilities.To_Upper
                     (Asis.Expressions.Name_Image (Element_Name));

            begin
               if Element_Image = "READ" then
                  Result.Read := Find_Property_Method (Element_Name);
                  if Declaration_Kind (Result.Read)
                    /= A_Function_Declaration
                  then
                     Amoc.Diagnoses.Raise_Exception
                      (Result.Read,
                       "Subprogram '"
                         & Name_Image (Result.Read)
                         & "' must be a function");
                  end if;
                  Result.Editable.Empty := True;
               end if;

               if Element_Image = "WRITE" then
                  Result.Write := Find_Property_Method (Element_Name);
               end if;

               if Element_Image = "RESET" then
                  Result.Reset := Find_Property_Method (Element_Name);
               end if;

               if Element_Image = "DESIGNABLE"
                 and Element_Name_Image = "TRUE"
               then
                     Result.Designable.State := True;
               end if;

               if Element_Image = "SCRIPTABLE"
                 and Element_Name_Image = "TRUE"
               then
                     Result.Scriptable.State := True;
               end if;

               if Element_Image = "STORED"
                 and Element_Name_Image = "TRUE" then
                     Result.Stored.State := True;
               end if;
            end;
         end loop;

         Result.Prop_Type := Result_Profile (Result.Read);

         if Is_Nil (Result.Prop_Type) then
            Amoc.Diagnoses.Raise_Exception
             (Result.Prop_Type,
              "Read operation '"
                & Name_Image (Result.Read)
                & "' must return value");
         end if;

         if not Is_Nil (Result.Write) then
            declare
               Params_List : constant Asis.Parameter_Specification_List
                  := Parameter_Profile (Result.Write);

            begin
               if Params_List'Length < 2 then
                  Amoc.Diagnoses.Raise_Exception
                   (Result.Write,
                    "Write operation subprogram must have two parameters.");
               end if;

               if Amoc.String_Utilities.To_Upper
                   (Asis.Expressions.Name_Image (Result.Prop_Type))
                  /= Amoc.String_Utilities.To_Upper
                      (Asis.Expressions.Name_Image
                        (Asis.Declarations.Object_Declaration_View
                          (Params_List (Params_List'Last))))
               then
                  Amoc.Diagnoses.Raise_Exception
                   (Result.Write,
                    "Type of the value must be same with the property's type");
               end if;
            end;
         end if;

         Result.Name := To_Unbounded_String
                         (Ada.Characters.Conversions.To_String
                           (Name_Image (Property_Element)));

         Properties.Append (Result);
      end;
   end Analyze_Pragma_Q_Property;

   -------------------------------
   -- Analyze_Type_Declarations --
   -------------------------------

   procedure Analyze_Type_Declarations
    (Visible_Declarations : Asis.Element_List)
   is
   begin
      for J in Visible_Declarations'Range loop
         if Asis.Elements.Declaration_Kind (Visible_Declarations (J))
              = Asis.A_Private_Extension_Declaration
         then
            --  Check is type derived from Q_Object_Director type.

            --  Ada semantic: it is impossible to declare tagged private
            --  extension without implementing all interfaces implemented by
            --  parent type, so ASIS implementation always check this rule
            --  and we just need to detect derivation from Q_Object_Director.

            if Amoc.Analizer.Types.Is_Derived_From_Q_Object_Director
                (Asis.Declarations.Corresponding_Type_Declaration
                  (Visible_Declarations (J)), False)
            then
               Tagged_Type_Declaration := Visible_Declarations (J);

               --  Output warning if type declaration in visible part also
               --  derived from Q_Object_Director. This is semantically
               --  possible, but makes available "protected" subprogram to
               --  accessed by everyone.

               if Amoc.Analizer.Types.Is_Derived_From_Q_Object_Director
                   (Visible_Declarations (J), True)
               then
                  Amoc.Diagnoses.Output
                   (Visible_Declarations (J),
                    Amoc.Diagnoses.
                      Type_Derived_From_Director_Type_In_Visible_Part);
               end if;

               exit;
            end if;
         end if;
      end loop;

      --  Output error if type derived from Q_Object_Director is not found.

      if Asis.Elements.Is_Nil (Tagged_Type_Declaration) then
         Amoc.Diagnoses.Output
          (Asis.Compilation_Units.Text_Name (Primary_Unit),
           1,
           1,
           Amoc.Diagnoses.Unable_To_Find_Type_Derived_From_Q_Object_Director);

         raise Amoc.Amoc_Error;
      end if;

      --  Determine name of the parent type's package and is it director or
      --  not.

      declare
         Aux : Asis.Element
           := Asis.Definitions.Subtype_Mark
               (Asis.Definitions.Parent_Subtype_Indication
                 (Asis.Declarations.Type_Declaration_View
                   (Asis.Declarations.Corresponding_Type_Declaration
                     (Tagged_Type_Declaration))));

      begin
         case Asis.Elements.Expression_Kind (Aux) is
            when Asis.A_Selected_Component =>
               Aux := Asis.Expressions.Selector (Aux);

            when Asis.An_Identifier =>
               null;

            when others =>
               raise Program_Error;
         end case;

         declare
            Unit                 : constant Asis.Compilation_Unit
              := Asis.Elements.Enclosing_Compilation_Unit
                  (Asis.Expressions.Corresponding_Name_Declaration (Aux));
            Visible_Declarations : constant Asis.Element_List
              := Asis.Declarations.Visible_Part_Declarative_Items
                  (Asis.Elements.Unit_Declaration (Unit));

         begin
            Parent_Package_Name :=
              +Asis.Compilation_Units.Unit_Full_Name (Unit);

            --  Searching for the declaration of the Static_Meta_Object.

            for J in Visible_Declarations'Range loop
               if Asis.Elements.Declaration_Kind (Visible_Declarations (J))
                    = Asis.A_Variable_Declaration
               then
                  declare
                     Names : constant Asis.Element_List
                       := Asis.Declarations.Names (Visible_Declarations (J));
                     Name  : constant Wide_String
                       := Asis.Declarations.Defining_Name_Image
                           (Names (Names'First));

                  begin
                     if Name = "Static_Meta_Object" then
                        Amoc.Model.Parent_Class_Is_Imported := True;

                        exit;
                     end if;
                  end;
               end if;
            end loop;
         end;
      end;
   end Analyze_Type_Declarations;

   -------------
   -- Cleanup --
   -------------

   procedure Cleanup is
   begin
      Amoc.Asis_Implementation.Cleanup_Asis_Context (Context);
   end Cleanup;

   -----------------------
   -- Compute_Signature --
   -----------------------

   function Compute_Signature
    (Subprogram_Name   : Wide_String;
     Parameters        : Amoc.Model.Parameter_Vectors.Vector;
     Requested_Name    : Wide_String;
     Requested_Profile : Wide_String)
       return Ada.Strings.Unbounded.Unbounded_String
   is
      Result   : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
      Is_Self  : Boolean := True;
      Is_First : Boolean := True;

      procedure Process_Parameter
       (Position : Amoc.Model.Parameter_Vectors.Cursor);

      -----------------------
      -- Process_Parameter --
      -----------------------

      procedure Process_Parameter
       (Position : Amoc.Model.Parameter_Vectors.Cursor)
      is
         Element : constant Amoc.Model.Parameter_Info
           := Amoc.Model.Parameter_Vectors.Element (Position);

      begin
         if Is_Self then
            Is_Self := False;

         else
            if Is_First then
               Is_First := False;

            else
               Ada.Strings.Wide_Unbounded.Append (Result, ',');
            end if;

            Ada.Strings.Wide_Unbounded.Append
             (Result,
              Amoc.Model.Known_Types.Element
               (Element.Type_Id).Qt_Name);

            if Element.Is_Access then
               Ada.Strings.Wide_Unbounded.Append (Result, "*");
            end if;
         end if;
      end Process_Parameter;

   begin
      if Requested_Name /= "" then
         Result :=
           Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
            (Requested_Name);

      else
         Result :=
           Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
            (To_Cpp_Method_Name (Subprogram_Name));
      end if;

      if Requested_Profile /= "" then
         Ada.Strings.Wide_Unbounded.Append (Result, Requested_Profile);

      else
         Ada.Strings.Wide_Unbounded.Append (Result, '(');

         Parameters.Iterate (Process_Parameter'Access);

         Ada.Strings.Wide_Unbounded.Append (Result, ')');
      end if;

      return
        Ada.Strings.Unbounded.To_Unbounded_String
         (Ada.Characters.Conversions.To_String
           (Ada.Strings.Wide_Unbounded.To_Wide_String (Result)));
   end Compute_Signature;

   -------------------------
   -- Extended_Name_Image --
   -------------------------

   function Extended_Name_Image (Element : Asis.Element)
     return Wide_String
   is
   begin
      case Asis.Elements.Expression_Kind (Element) is
         when Asis.An_Identifier =>
            return Asis.Expressions.Name_Image (Element);

         when Asis.A_Selected_Component =>
            return
              Extended_Name_Image (Asis.Expressions.Prefix (Element))
                & "."
                & Extended_Name_Image
                   (Asis.Expressions.Selector (Element));

         when others =>
            raise Program_Error;
      end case;
   end Extended_Name_Image;

   -------------
   -- Prepare --
   -------------

   procedure Prepare is
   begin
      Amoc.Asis_Implementation.Prepare_Asis_Context (Context);

      --  Search for primary compilation unit

      Primary_Unit :=
        Asis.Compilation_Units.Library_Unit_Declaration
         (Amoc.Model.Compilation_Unit_Name, Context);

      if Asis.Compilation_Units.Is_Nil (Primary_Unit) then
         Amoc.Diagnoses.Output
          (Amoc.Parameters.Source_File_Name,
           1,
           1,
           "expected compilation unit not found in the specified source file");

         raise Amoc.Amoc_Error;
      end if;

      --  Process known Qt types

      Amoc.Analizer.Types.Process_Qt_Types;
   end Prepare;

   ------------------------
   -- To_Cpp_Method_Name --
   ------------------------

   function To_Cpp_Method_Name (Name : Wide_String) return Wide_String is
      Result      : Wide_String (1 .. Name'Length);
      Last        : Natural := 0;
      Force_Upper : Boolean := False;

   begin
      for J in Name'Range loop
         if J = Name'First then
            Last := Last + 1;
            Result (Last) := Amoc.String_Utilities.To_Lower (Name (J));

         elsif Name (J) = '_' then
            Force_Upper := True;

         else
            Last := Last + 1;

            if Force_Upper then
               Result (Last) := Amoc.String_Utilities.To_Upper (Name (J));
               Force_Upper := False;

            else
               Result (Last) := Name (J);
            end if;
         end if;
      end loop;

      return Result (Result'First .. Last);
   end To_Cpp_Method_Name;

end Amoc.Analizer;
