------------------------------------------------------------------------------
--                                                                          --
--                An Ada2005 binding to the Qt C++ framework                --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2008, 2009 Vadim Godunko <vgodunko@gmail.com>                --
--                                                                          --
-- QtAda is free software;  you can  redistribute it and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. QtAda is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License distributed with QtAda;  see file COPYING.  If not, write --
-- to  the  Free Software Foundation,  51  Franklin  Street,  Fifth  Floor, --
-- Boston, MA 02110-1301, USA.                                              --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 3321 $ $Date: 2009-05-04 01:05:25 +0400 (Mon, 04 May 2009) $
------------------------------------------------------------------------------
with Ada.Characters.Conversions;
with Ada.Directories;
with Ada.Wide_Text_IO;

with Asis.Compilation_Units;
with Asis.Elements;
with Asis.Text;

package body Amoc.Diagnoses is

   type Wide_String_Access is access all Wide_String;

   Message_Text : constant array (Message_Kind) of Wide_String_Access
     := (Duplicate_Pragma_Q_Object_Declaration =>
           new Wide_String'("Duplicate pragma Q_Object declaration"),
         First_Argument_Pragma_Q_Object_Must_Be_Identifier =>
           new Wide_String'
                ("first argument of pragma Q_Object must be direct"
                   & " identifier"),
         First_Argument_Pragma_Q_Object_Must_Be_Type_Name =>
           new Wide_String'
                ("first argument of pragma Q_Object must be tagged type's"
                   & " name"),
         Pragma_Q_Object_Must_Have_2_Arguments =>
           new Wide_String'("pragma Q_Object must have two arguments"),
         Q_Signal_1st_Must_Be_Direct_Name =>
           new Wide_String'
                ("first parameter of pragma Q_Signal must be direct name"
                   & " of the signal's subprogram"),
         Q_Signal_2nd_Must_Be_String_Literal =>
           new Wide_String'
                ("second parameter of pragma Q_Signal must be string literal"),
         Q_Signal_Must_Have_1_Or_2_Parameters =>
           new Wide_String'("pragma Q_Signal must have one or two parameters"),
         Q_Signal_Subprogram_Must_Be_Procedure =>
           new Wide_String'("subprogram for signal must be a procedure"),
         Q_Signal_Subprogram_Must_Start_From_Emit =>
           new Wide_String'
                ("name of the subprogram for signal must start from 'Emit_'"),
         Q_Signal_Subprogram_Not_Found =>
           new Wide_String'
                ("subprogram specified in pragma Q_Signal is not found"),
         Q_Slot_Must_Have_1_Or_2_Parameters =>
           new Wide_String'("pragma Q_Slot must have one or two parameters"),
         Q_Slot_1st_Must_Be_Direct_Name =>
           new Wide_String'
                ("first parameter of pragma Q_Slot must be direct name"
                   & " of the slot's subprogram"),
         Q_Slot_2nd_Must_Be_String_Literal =>
           new Wide_String'
                ("second parameter of pragma Q_Slot must be string literal"),
         Q_Slot_Subprogram_Not_Found =>
           new Wide_String'
                ("subprogram specified in pragma Q_Slot is not found"),
         Q_Invokable_Must_Have_1_Or_2_Parameters =>
           new Wide_String'
                ("pragma Q_Invokable must have one or two parameters"),
         Q_Invokable_1st_Must_Be_Direct_Name =>
           new Wide_String'
                ("first parameter of pragma Q_Invokable must be direct name"
                   & " of the operation's subprogram"),
         Q_Invokable_2nd_Must_Be_String_Literal =>
           new Wide_String'
                ("second parameter of pragma Q_Invokable must be string"
                   & " literal"),
         Q_Invokable_Subprogram_Not_Found =>
           new Wide_String'
                ("subprogram specified in pragma Q_Invokable is not found"),
         Second_Argument_Pragma_Q_Object_Must_Be_String =>
           new Wide_String'
                ("second argument of pragma Q_Object must be string literal"),
         Second_Argument_Pragma_Q_Object_Must_Be_Valid =>
           new Wide_String'
                ("second argument of pragma Q_Object must be string literal"
                   & " which contains valid C++ identifier"),
         Type_Derived_From_Director_Type_In_Visible_Part =>
           new Wide_String'
                ("warning: type derived from director type in visible part"),
         Unable_To_Compile_Source_File =>
           new Wide_String'
                ("unable to compile source file into the GNAT Ada Tree file"),
         Unable_To_Find_Type_Derived_From_Q_Object_Director =>
           new Wide_String'
                ("unable to find tagged type derived from the QObject"
                   & " director"),
         Unknown_Pragma =>
           new Wide_String'("warning: unknown pragma"));

   function Trimmed_Image (Item : Integer) return Wide_String;
   --  Equivalent to Integer'Wide_Image except all spaces are trimmed out.

   ------------
   -- Output --
   ------------

   procedure Output
    (File_Name : Wide_String;
     Line      : Positive;
     Column    : Positive;
     Message   : Wide_String)
   is
      Line_Image   : constant Wide_String := Trimmed_Image (Line);
      Column_Image : constant Wide_String := Trimmed_Image (Column);
      File_Image   : constant Wide_String
        := Ada.Characters.Conversions.To_Wide_String
            (Ada.Directories.Simple_Name
              (Ada.Characters.Conversions.To_String (File_Name)));

   begin
      Ada.Wide_Text_IO.Put_Line
       (Ada.Wide_Text_IO.Standard_Error,
        File_Image & ':' & Line_Image & ':' & Column_Image & ": " & Message);
   end Output;

   ------------
   -- Output --
   ------------

   procedure Output
    (File_Name : Wide_String;
     Line      : Positive;
     Column    : Positive;
     Message   : Message_Kind)
   is
   begin
      Output (File_Name, Line, Column, Message_Text (Message).all);
   end Output;

   ------------
   -- Output --
   ------------

   procedure Output
    (Element : Asis.Element;
     Message : Message_Kind)
   is
      Span : constant Asis.Text.Span := Asis.Text.Element_Span (Element);

   begin
      Output
       (Asis.Compilation_Units.Text_Name
         (Asis.Elements.Enclosing_Compilation_Unit (Element)),
        Positive (Span.First_Line),
        Positive (Span.First_Column),
        Message_Text (Message).all);
   end Output;

   ---------------------
   -- Raise_Exception --
   ---------------------

   procedure Raise_Exception (Item : Asis.Element; Text : Wide_String)
   is
   begin
      Ada.Wide_Text_IO.Put_Line (Text);
      Ada.Wide_Text_IO.Put_Line (Asis.Elements.Debug_Image (Item));

      raise Program_Error;
   end Raise_Exception;

   -------------------
   -- Trimmed_Image --
   -------------------

   function Trimmed_Image (Item : Integer) return Wide_String is
      Image : constant Wide_String := Integer'Wide_Image (Item);

   begin
      if Image (Image'First) = ' ' then
         return Image (Image'First + 1 .. Image'Last);

      else
         return Image;
      end if;
   end Trimmed_Image;

end Amoc.Diagnoses;
