------------------------------------------------------------------------------
--                                                                          --
--                An Ada2005 binding to the Qt C++ framework                --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2008-2010 Vadim Godunko <vgodunko@gmail.com>                 --
--                                                                          --
-- QtAda is free software;  you can  redistribute it and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. QtAda is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License distributed with QtAda;  see file COPYING.  If not, write --
-- to  the  Free Software Foundation,  51  Franklin  Street,  Fifth  Floor, --
-- Boston, MA 02110-1301, USA.                                              --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 3792 $ $Date: 2010-10-29 20:33:40 +0400 (Fri, 29 Oct 2010) $
------------------------------------------------------------------------------
with Ada.Characters.Conversions;
with Ada.Characters.Handling;
with Ada.Containers;
with Ada.Directories;
with Ada.Integer_Text_IO;
with Ada.Strings.Wide_Unbounded;
with Ada.Text_IO;
with Interfaces;

with Asis.Declarations;
with Asis.Elements;
with Asis.Expressions;

with Amoc.Generator.Dependencies;
with Amoc.Model;
with Amoc.Parameters;
with Amoc.Plugin_Verification;

package body Amoc.Generator is

   use Ada.Containers;
   use Ada.Integer_Text_IO;
   use Ada.Text_IO;
   use Amoc.Model;
   use Amoc.Model.Method_Info_Vectors;
   use Amoc.Model.Declared_Interface_Info_Vectors;
   use Amoc.Model.Property_Info_Vectors;
   use Interfaces;
   use type Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;

   procedure Generate_Meta_Data;
   --  Производит генерацию файла метаинформации.

   procedure Generate_Meta_Call_Implementation (File : in out File_Type);
   --  Производит генерацию реализации подпрограммы вызова мета операции.

   procedure Generate_Meta_Cast_Implementation (File : in out File_Type);
   --  Производит генерацию файла с реализацией функции приведения указателей.

   procedure Generate_Signals_Implementation;
   --  Производит генерацию реализации сигналов.

   function "-" (Item : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String)
     return String;

   type Flags_Type is new Interfaces.Unsigned_32;

   package Flags_Type_IO is
     new Ada.Text_IO.Modular_IO (Flags_Type);

   ---------
   -- "-" --
   ---------

   function "-" (Item : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String)
     return String
   is
   begin
      return
        Ada.Characters.Conversions.To_String
         (Ada.Strings.Wide_Unbounded.To_Wide_String (Item));
   end "-";

   --------------
   -- Generate --
   --------------

   procedure Generate is
   begin
      Generate_Meta_Data;
      Generate_Signals_Implementation;

      if Amoc.Parameters.Output_Dependencies then
         Amoc.Generator.Dependencies.Generate;
      end if;
   end Generate;

   ---------------------------------------
   -- Generate_Meta_Call_Implementation --
   ---------------------------------------

   procedure Generate_Meta_Call_Implementation (File : in out File_Type) is
      Method_Index : Natural := 0;

      procedure Generate_Method_Call
       (Position : Amoc.Model.Method_Info_Vectors.Cursor);
      procedure Generate_Property_Method;

      --------------------------
      -- Generate_Method_Call --
      --------------------------

      procedure Generate_Method_Call
       (Position : Amoc.Model.Method_Info_Vectors.Cursor)
      is
         Method : constant Amoc.Model.Method_Info
           := Amoc.Model.Method_Info_Vectors.Element (Position);
         Args   : constant Amoc.Model.Parameter_Vectors.Vector
           := Method.Parameters;

      begin
         Ada.Text_IO.Put_Line
          (File,
           "            when"
             & Integer'Image (Method_Index)
             & " =>");

         if Args.Length = 1 then
            Ada.Text_IO.Put_Line
             (File,
              "               "
                & Ada.Strings.Unbounded.To_String (Method.Subprogram_Name));
            Ada.Text_IO.Put_Line
             (File,
              "                ("
                & (-Amoc.Model.Tagged_Type_Name)
                & "'Class (Self.all)'Access);");

         else
            Ada.Text_IO.Put_Line (File, "               declare");
            Ada.Text_IO.Put_Line
             (File,
              "                  Arg_Array : constant Qt4.Address_Array");
            Ada.Text_IO.Put_Line
             (File,
              "                    := "
                & "Qt4.Address_Pointers.Value "
                & "(Qt4.Address_Pointers.Pointer (Args),"
                & Integer'Image (Integer (Args.Length))
                &");");
            Ada.Text_IO.New_Line (File);

            for J in 2 .. Integer (Args.Length) loop
               declare
                  Num        : constant String := Integer'Image (J - 1);
                  Param_Type : constant Amoc.Model.Type_Info
                    := Amoc.Model.Known_Types.Element
                        (Args.Element (J).Type_Id);

               begin
                  if Param_Type.Kind = Amoc.Model.Primitive
                    or else Param_Type.Kind = Amoc.Model.Enumeration
                  then
                     Put_Line
                      (File,
                       "                  package Convert_"
                         & Num (Num'First + 1 .. Num'Last)
                         & " is");
                     Put_Line
                      (File,
                       "                    new System."
                         & "Address_To_Access_Conversions");
                     Put (File, "                         (");
                     Put
                      (File,
                       Ada.Characters.Conversions.To_String
                        (Ada.Strings.Wide_Unbounded.To_Wide_String
                          (Param_Type.Ada_Name)));
                     Put_Line (File, ");");
                  end if;
               end;
            end loop;

            Ada.Text_IO.New_Line (File);
            Ada.Text_IO.Put_Line (File, "               begin");
            Ada.Text_IO.Put_Line
             (File,
              "                  "
                & Ada.Strings.Unbounded.To_String (Method.Subprogram_Name));
            Ada.Text_IO.Put
             (File,
              "                   ("
                & (-Amoc.Model.Tagged_Type_Name)
                & "'Class (Self.all)'Access");

            for J in 2 .. Integer (Args.Length) loop
               declare
                  Num   : constant String := Integer'Image (J - 1);
                  Value : constant String
                    := Num (Num'First + 1 .. Num'Last);
                  Param_Type : constant Amoc.Model.Type_Info
                    := Amoc.Model.Known_Types.Element
                        (Args.Element (J).Type_Id);

               begin
                  Put_Line (File, ",");

                  case Param_Type.Kind is
                     when Amoc.Model.Primitive
                       | Amoc.Model.Enumeration
                     =>
                        Put (File,
                             "Convert_"
                               & Value
                               & ".To_Pointer (Arg_Array ("
                               & "Arg_Array'First + "
                               & Value
                               & ")).all");

                     when Amoc.Model.Value_Class =>
                        Put_Line
                         (File,
                          "                    "
                            & Ada.Characters.Conversions.To_String
                               (Ada.Strings.Wide_Unbounded.To_Wide_String
                                 (Param_Type.Create_Wrapper)));
                        Put
                         (File,
                          "                     ("
                            & Ada.Characters.Conversions.To_String
                               (Ada.Strings.Wide_Unbounded.To_Wide_String
                                 (Param_Type.To_Pointer))
                            & " (Arg_Array (Arg_Array'First + "
                            & Value
                            & ")))");

                     when Amoc.Model.QObject_Subclass =>
                        Put
                         (File,
                          "                    "
                            & Ada.Characters.Conversions.To_String
                               (Ada.Strings.Wide_Unbounded.To_Wide_String
                                 (Param_Type.To_Pointer))
                            & " (Arg_Array (Arg_Array'First + "
                            & Value
                            & "))");
                  end case;
               end;
            end loop;

            Put_Line (File, ");");
            Put_Line (File, "               end;");
         end if;

         New_Line (File);

         Method_Index := Method_Index + 1;
      end Generate_Method_Call;

      ------------------------------
      -- Generate_Property_Method --
      ------------------------------

      procedure Generate_Property_Method is
         Ind : Natural := 0;

         procedure Generate_End;

         procedure Genereate_P (E : Asis.Element; T : Asis.Element);

         ------------------
         -- Generate_End --
         ------------------

         procedure Generate_End is
         begin
            New_Line (File);
            Put_Line (File, "         when others =>");
            Put_Line (File, "            null;");
            Put_Line (File, "      end case;");
            Put_Line (File, "      XYZ_Id := Qt4.""-"" (XYZ_Id, "
                              & Integer'Image (Integer (Length (Properties)))
                              & ");");
            Put_Line (File, "   end if;");
            New_Line (File);
         end Generate_End;

         -----------------
         -- Genereate_P --
         -----------------

         procedure Genereate_P (E : Asis.Element; T : Asis.Element) is
            use type Asis.Declaration_Kinds;
            use type Asis.Element_Kinds;

         begin
            if not (Asis.Elements.Is_Nil (E)) then
               declare
                  Name  : constant Asis.Element_List
                    := Asis.Declarations.Names (E);
                  Image : constant String
                    := Ada.Characters.Conversions.To_String
                         (Asis.Declarations.Defining_Name_Image
                           (Name (Name'First)));
                  Args : constant Asis.Element_List
                    := Asis.Declarations.Parameter_Profile (E);

                  Type_Image : constant String
                    := Ada.Characters.Conversions.To_String
                       (Asis.Expressions.Name_Image (T));

               begin
                  New_Line (File);
                  Put_Line (File, "         when" & Integer'Image (Ind)
                                   & " =>");
                  Put_Line (File, "            declare");
                  Put_Line (File, "               Arg_Array : constant"
                                    & " Qt4.Address_Array");
                  Put_Line (File, "                 := "
                                    & "Qt4.Address_Pointers.Value ");
                  Put (File, "                     ");
                  Put_Line (File,  "(Qt4.Address_Pointers.Pointer (Args),"
                                    & Integer'Image (Args'Length)
                                    &");");
                  Put_Line (File,
                            "               package Convert_" & Image & " is");
                  Put_Line (File, "                 "
                                    & "new "
                                    & "System.Address_To_Access_Conversions");
                  Put_Line (File, "                      ("
                                    & Type_Image & ");");
                  New_Line (File);
                  Put_Line (File, "            begin");

                  if Asis.Elements.Element_Kind (E) = Asis.A_Declaration then
                     if Asis.Elements.Declaration_Kind (E)
                         = Asis.A_Function_Declaration
                     then
                        Put_Line (File, "               Convert_"
                                         & Image & ".To_Pointer");
                        Put_Line (File,
                                  "                "
                                   & "(Arg_Array (Arg_Array'First)).all := ");
                        Put_Line (File, "                  "
                                         & Image & "(Self);");
                        Put_Line (File, "            end;");

                     elsif Asis.Elements.Declaration_Kind (E)
                             = Asis.A_Procedure_Declaration
                     then
                        Put_Line (File, "               " & Image);
                        Put_Line (File, "                (Self,");
                        Put_Line (File, "                 Convert_" & Image
                                          & ".To_Pointer ");
                        Put_Line (File, "                  (Arg_Array"
                                          & "(Arg_Array'First + 1)).all);");
                        Put_Line (File, "            end;");
                     end if;
                  end if;
               end;
               New_Line (File);
            end if;
            end Genereate_P;

      begin

         Put_Line (File, "   if Kind = Qt4.Meta_Objects.Read_Property"
                           & " then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Read,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects.Write_Property"
                           & " then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Write,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects.Reset_Property"
                           & " then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects."
                           & "Query_Property_Designable then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects."
                           & "Query_Property_Scriptable then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects."
                           & "Query_Property_Stored then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects."
                           & "Query_Property_Editable then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
         Ind := 0;
         Put_Line (File, "   if Kind = Qt4.Meta_Objects."
                           & "Query_Property_User then");
         Put_Line (File, "      case XYZ_Id is");

         for J in 1 .. Integer (Length (Properties)) loop
            Genereate_P (Element (Properties, J).Reset,
                         Element (Properties, J).Prop_Type);
            Ind := Ind + 1;
         end loop;

         Generate_End;
      end Generate_Property_Method;

------------------------------------------------------------------------------
--   Пример обработки процедуры с параметрами.
--  declare
--    Arg_Array : constant Qt4.Address_Array
--      := Qt4.Address_Pointers.Value (Args);
--
--  package Convert_2 is new System.Address_To_Access_Conversions (My_Type);
--  package Convert_3 is new System.Address_To_Access_Conversions (My_Access);
--
--  begin
--    My_Method (Self,
--               Convert_2.To_Pointer (Arg_Array (2)).all,
--               Convert_3.To_Pointer (Arg_Array (3)));
--  and;
------------------------------------------------------------------------------

   begin
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "   ------------------");
      Ada.Text_IO.Put_Line (File, "   -- Qt_Meta_Call --");
      Ada.Text_IO.Put_Line (File, "   ------------------");
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "   function Qt_Meta_Call");
      Ada.Text_IO.Put_Line
       (File,
        "    (Self : not null"
          & " Qt_Ada.Internals.Objects.Q_Object_Base_Access;");
      Ada.Text_IO.Put_Line (File, "     Kind : Qt4.Meta_Objects.Call;");
      Ada.Text_IO.Put_Line (File, "     Id   : Qt4.Q_Integer;");
      Ada.Text_IO.Put_Line (File, "     Args : Qt4.Address_Pointer)");
      Ada.Text_IO.Put_Line (File, "       return Qt4.Q_Integer");
      Ada.Text_IO.Put_Line (File, "   is");
      Ada.Text_IO.Put_Line (File, "      use type Qt4.Meta_Objects.Call;");
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "      XYZ_Id : Qt4.Q_Integer;");
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "   begin");
      Ada.Text_IO.Put_Line (File, "      XYZ_Id := Id;");

      if not Amoc.Model.Signals.Is_Empty
        or else not Amoc.Model.Slots.Is_Empty
        or else not Amoc.Model.Methods.Is_Empty
      then
         New_Line (File);
         Put_Line (File, "      if Kind = Qt4.Meta_Objects.Invoke_Meta_Method"
                           & " then");
         Put_Line (File, "         case XYZ_Id is");

         Amoc.Model.Signals.Iterate (Generate_Method_Call'Access);
         Amoc.Model.Slots.Iterate (Generate_Method_Call'Access);
         Amoc.Model.Methods.Iterate (Generate_Method_Call'Access);

         Put_Line (File, "            when others =>");
         Put_Line (File, "               null;");
         Put_Line (File, "         end case;");
         New_Line (File);
         Put_Line (File, "         XYZ_Id := Qt4.""-"" (XYZ_Id, "
                           & Integer'Image
                              (Integer (Amoc.Model.Signals.Length)
                                 + Integer (Amoc.Model.Slots.Length)
                                 + Integer (Amoc.Model.Methods.Length))
                           & ");");

         Put_Line (File, "      end if;");
      end if;

      if Length (Properties) /= 0 then
         New_Line (File);
         Generate_Property_Method;
      end if;

      New_Line (File);
      Put_Line (File, "      return XYZ_Id;");
      Put_Line (File, "   end Qt_Meta_Call;");
   end Generate_Meta_Call_Implementation;

   ---------------------------------------
   -- Generate_Meta_Cast_Implementation --
   ---------------------------------------

   procedure Generate_Meta_Cast_Implementation (File : in out File_Type) is
   begin
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "   ------------------");
      Ada.Text_IO.Put_Line (File, "   -- Qt_Meta_Cast --");
      Ada.Text_IO.Put_Line (File, "   ------------------");
      Ada.Text_IO.New_Line (File);
      Ada.Text_IO.Put_Line (File, "   function Qt_Meta_Cast");
      Ada.Text_IO.Put_Line
       (File,
        "    (Self  : not null"
          & " Qt_Ada.Internals.Objects.Q_Object_Base_Access;");
      Ada.Text_IO.Put_Line (File, "     Class : String)");
      Ada.Text_IO.Put_Line (File, "       return System.Address");
      Ada.Text_IO.Put_Line (File, "   is");
      Ada.Text_IO.Put_Line (File, "   begin");

      Put_Line (File, "      if Class = """ & (-Amoc.Model.Custom_Class_Name)
                        & """ then");
      Put_Line
       (File,
        "         return Qt_Ada.Internals.Objects.Conversions.To_Address");
      Put_Line
       (File, "                 (Qt_Ada.Internals.Objects.QObject_Access");
      Put_Line (File, "                   (Self.QObject_View));");
      New_Line (File);

      --  Генерация сравнений с интерфейсами указанными в
      --  pragma Q_Declare_Interface.

      for J in 1 .. Integer (Length (Declared_Interfaces)) loop
--          Put_Line (File, "   elsif Class = """
--                     & To_String (Declared_Interface_Info_Vectors.Element
--                                   (Declared_Interfaces, J).Name)
--                     & """ then");
--        Put_Line (File, "      return Qt4_Entities.Conversions.To_Address");
--          Put_Line (File, "       (Qt4_Entities.Conversions.Object_Pointer "
--                            & "(Self.Cpp_Object));");
--          New_Line (File);

         Put_Line (File, "      elsif Class = "
                    & To_String (Declared_Interface_Info_Vectors.Element
                                  (Declared_Interfaces, J).Identifier)
                    & " then");

         if To_String (Declared_Interface_Info_Vectors.Element
                                  (Declared_Interfaces, J).Identifier)
           = '"' & "com.trolltech.Qt.Designer.CustomWidget" & '"'
         then
            Put_Line
             (File,
              "         return "
                & (-Amoc.Model.Tagged_Type_Name)
                & " (Self.all)."
                & "Q_Designer_Custom_Widget_Interface_View.Cpp_Object;");

         else
            Put_Line (File, "         return");
            Put_Line (File, "           Qt_Ada.Internals.Objects.Conversions."
                              & "To_Address");
            Put_Line
             (File, "            (Qt_Ada.Internals.Objects.QObject_Access");
            Put_Line (File, "              (Self.QObject_View));");
         end if;

         New_Line (File);
      end loop;

      Put_Line (File, "      else");
      Put_Line (File, "         return System.Null_Address;");
      Put_Line (File, "      end if;");

      Put_Line (File, "   end Qt_Meta_Cast;");
   end Generate_Meta_Cast_Implementation;

   ------------------------
   -- Generate_Meta_Data --
   ------------------------

   procedure Generate_Meta_Data is
      File : File_Type;
      Name : constant String
        := Ada.Directories.Base_Name
            (Ada.Characters.Conversions.To_String
              (Amoc.Parameters.Source_File_Name));

      Data_Offset         : Natural := 12;
      --  First unused position in the data array. First 12 elements have
      --  special purpose - they describe below layout of the data in the
      --  array.

      String_Data_Offset  : Natural
        := Ada.Strings.Wide_Unbounded.Length (Amoc.Model.Custom_Class_Name)
             + 2;
      --  Текущее неиспользованное смещение в массиве строковых данных.

      Empty_String_Offset : constant Natural
        := Ada.Strings.Wide_Unbounded.Length (Amoc.Model.Custom_Class_Name)
            + 1;
      --  Смещение "пустой" строки.

      function Method_Properties (Item : Method_Info) return Integer;
      --  Формирует значение свойств метода в метасистеме.

      procedure Generate_Data;
      --  Производит генерацию кода массива числовых данных.

      function Property_Flags (Property : Property_Info) return Flags_Type;
      --  Формирует флаги для свойств

      procedure Generate_String_Data;
      --  Производит генерацию кода массива строковых данных.

      procedure Generate_Context_Clauses
       (Position : Amoc.Model.Method_Info_Vectors.Cursor);
      --  Generate context clauses for utilities package for used QObject
      --  subclasses.

      ------------------------------
      -- Generate_Context_Clauses --
      ------------------------------

      procedure Generate_Context_Clauses
       (Position : Amoc.Model.Method_Info_Vectors.Cursor)
      is
         Element : constant Amoc.Model.Method_Info
           := Amoc.Model.Method_Info_Vectors.Element (Position);
         Args : constant Amoc.Model.Parameter_Vectors.Vector
           := Element.Parameters;

      begin
         for J in 2 .. Integer (Args.Length) loop
            declare
               Param_Type : constant Amoc.Model.Type_Info
                 := Amoc.Model.Known_Types.Element
                     (Args.Element (J).Type_Id);

            begin
               if Param_Type.Kind = Amoc.Model.QObject_Subclass then
                  Ada.Text_IO.Put_Line
                   (File,
                    "with "
                      & Ada.Characters.Conversions.To_String
                         (Ada.Strings.Wide_Unbounded.To_Wide_String
                           (Param_Type.Utils_Package))
                      & ";");
               end if;
            end;
         end loop;
      end Generate_Context_Clauses;

      -------------------
      -- Generate_Data --
      -------------------

      procedure Generate_Data is

         procedure Generate_Methods
          (Kind    : String;
           Methods : Method_Info_Vectors.Vector);
         --  Производит генерацию данных для методов.

         procedure Generate_Properties
         (Properties : Property_Info_Vectors.Vector);
         --  Производит генерацию данных для свойств,
         --  объявленных при помощи Q_Property.

         procedure Generate_Methods
          (Kind    : String;
           Methods : Method_Info_Vectors.Vector)
         is
         begin
            if Length (Methods) = 0 then
               return;
            end if;

            Put_Line (File, "         --  "
                              & Kind
                              & ": signature, parameters, type, tag, flags");

            for J in 1 .. Integer (Length (Methods)) loop
               Put (File, "         ");
               Put (File, String_Data_Offset);
               Put (File, ',');
               Put (File, Empty_String_Offset);
               Put (File, ',');
               Put (File, Empty_String_Offset);
               Put (File, ',');
               Put (File, Empty_String_Offset);
               Put (File, ',');
               Put (File,
                    Method_Properties (Element (Methods, J)), Base => 16);
               Put_Line (File, ",");

               String_Data_Offset
                 := String_Data_Offset
                     + Length (Element (Methods, J).Signature)
                     + 1;
            end loop;
         end Generate_Methods;

         -------------------------
         -- Generate_Properties --
         -------------------------

         procedure Generate_Properties
          (Properties : Property_Info_Vectors.Vector)
         is
            Type_String_Data_Offset : Natural := String_Data_Offset;
         begin
            if Length (Properties) = 0 then
               return;
            end if;

            Put_Line (File, "         --  properties: name, type, flags");

            for J in 1 .. Integer (Length (Properties)) loop
               Put (File, "         ");
               String_Data_Offset :=
                 Type_String_Data_Offset
                   + Asis.Expressions.Name_Image
                      (Element (Properties, J).Prop_Type)'Length
                   + 1;
               Put (File, String_Data_Offset);

               Put (File, ',');
               Put (File, Type_String_Data_Offset);
               Type_String_Data_Offset :=
                 String_Data_Offset
                   + To_String (Element (Properties, J).Name)'Length
                   + 1;
               Put (File, ",");
               Put (File, "         ");
               Flags_Type_IO.Put
                 (File,
                  Property_Flags (Element (Properties, J)), Base => 16);
               Put_Line (File, ",");
            end loop;
         end Generate_Properties;

      begin
         Put_Line (File, "   Data : constant array (Positive range <>)"
                           & " of aliased Qt4.Q_Unsigned_Integer");
         Put (File, "     := (");
         Put (File, "          2");
         Put_Line (File, ",              --  Revision");
         Put (File, "         ");
         Put (File, 0);
         Put_Line (File, ",              --  Class Name");
         Put (File, "         ");
         Put (File, 0);
         Put (File, ',');
         Put (File, 0);
         Put_Line (File, ",  --  Class Info");
         Put (File, "         ");
         Put
          (File,
           Integer (Amoc.Model.Signals.Length)
             + Integer (Amoc.Model.Slots.Length)
             + Integer (Amoc.Model.Methods.Length));
         Put (File, ',');

         if not Amoc.Model.Signals.Is_Empty
           or else not Amoc.Model.Slots.Is_Empty
           or else not Amoc.Model.Methods.Is_Empty
         then
            Put (File, Data_Offset);

         else
            Put (File, 0);
         end if;

         Put_Line (File, ",  --  Methods");

         Data_Offset :=
           Data_Offset
             + (Integer (Amoc.Model.Signals.Length)
                  + Integer (Amoc.Model.Slots.Length)
                  + Integer (Amoc.Model.Methods.Length)) * 5;

         Put (File, "         ");
         Put (File, Integer (Length (Properties)));
         Put (File, ',');

         if Length (Properties) > 0 then
            Put (File, Data_Offset);

         else
            Put (File, 0);
         end if;

         Data_Offset := Data_Offset + Integer (Length (Properties)) * 3;

         Put_Line (File, ",  --  Properties");
         Put (File, "         ");
         Put (File, 0);
         Put (File, ',');
         Put (File, 0);
         Put_Line (File, ",  --  Enumerations/Sets");
         Put (File, "         ");
         Put (File, 0);
         Put (File, ',');
         Put (File, 0);
         Put_Line (File, ",  --  Constructors");

         Generate_Methods ("Signals", Signals);
         Generate_Methods ("Slots", Slots);
         Generate_Methods ("Methods", Methods);
         Generate_Properties (Properties);

         Put (File, "         ");
         Put (File, 0);
         Put_Line (File, ");             --  End of data");
      end Generate_Data;

      --------------------------
      -- Generate_String_Data --
      --------------------------

      procedure Generate_String_Data is

         procedure Generate_Method_Signature_String_Data
          (Position : Amoc.Model.Method_Info_Vectors.Cursor);

         -------------------------------------------
         -- Generate_Method_Signature_String_Data --
         -------------------------------------------

         procedure Generate_Method_Signature_String_Data
          (Position : Amoc.Model.Method_Info_Vectors.Cursor)
         is
            Element : constant Amoc.Model.Method_Info
              := Amoc.Model.Method_Info_Vectors.Element (Position);

         begin
            Ada.Text_IO.Put_Line
             (File,
              "          & """
                & To_String (Element.Signature)
                & """ & Interfaces.C.nul");
         end Generate_Method_Signature_String_Data;

      begin
         Put_Line (File, "   String_Data : constant Interfaces.C.char_array");

         Put
          (File,
           "     := """
             & (-Amoc.Model.Custom_Class_Name)
             & """ & Interfaces.C.nul");

         if not Amoc.Model.Signals.Is_Empty
           or else not Amoc.Model.Slots.Is_Empty
           or else not Amoc.Model.Methods.Is_Empty
         then
            Put_Line (File, " & Interfaces.C.nul");

         else
            New_Line;
         end if;

         Amoc.Model.Signals.Iterate
          (Generate_Method_Signature_String_Data'Access);
         Amoc.Model.Slots.Iterate
          (Generate_Method_Signature_String_Data'Access);
         Amoc.Model.Methods.Iterate
          (Generate_Method_Signature_String_Data'Access);

         for J in 1 .. Integer (Length (Properties)) loop
            Put_Line (File, "          & """
                              & Ada.Characters.Conversions.To_String
                                 (Asis.Expressions.Name_Image
                                  (Element (Properties, J).Prop_Type)
                              & """ & Interfaces.C.nul"));
            Put_Line (File, "          & """
                              & To_String (Element (Properties, J).Name)
                              & """ & Interfaces.C.nul");
         end loop;

         Put_Line (File, ";");
      end Generate_String_Data;

      -----------------------
      -- Method_Properties --
      -----------------------

      function Method_Properties (Item : Method_Info) return Integer is
         Result : Integer := 0;

         --  Константы должны быть согласованы с компилятором moc из состава
         --  Qt. Для версии Qt 4.5.0 это src/tools/moc/generator.cpp.

         --  Access:
         --    AccessPrivate   = 0x00
         --    AccessProtected = 0x01
         --    AccessPublic    = 0x02
         Access_Private   : constant := 16#00#;
         Access_Protected : constant := 16#01#;
         Access_Public    : constant := 16#02#;

         --  Method
         --    MethodMethod        = 0x00
         --    MethodSignal        = 0x04
         --    MethodSlot          = 0x08
         --    MethodConstructor   = 0x0c
         --    MethodCompatibility = 0x10
         --    MethodCloned        = 0x20
         --    MethodScriptable    = 0x40

         Method_Method        : constant := 16#00#;
         Method_Signal        : constant := 16#04#;
         Method_Slot          : constant := 16#08#;
--       Method_Constructor   : constant := 16#0C#;
--       Method_Compatibility : constant := 16#10#;
--       Method_Cloned        : constant := 16#20#;
--       Method_Scriptable    : constant := 16#40#;

      begin
         case Item.Kind is
            when Method =>
               Result := Result + Method_Method;

            when Signal =>
               Result := Result + Method_Signal;

            when Slot =>
               Result := Result + Method_Slot;
         end case;

         case Item.Visibility is
            when Amoc.Model.A_Public =>
               Result := Result + Access_Public;

            when Amoc.Model.A_Protected =>
               Result := Result + Access_Protected;

            when Amoc.Model.A_Private =>
               Result := Result + Access_Private;
         end case;

         return Result;
      end Method_Properties;

      --------------------
      -- Property_Flags --
      --------------------

      function Property_Flags (Property : Property_Info) return Flags_Type is

      --  Константы должны быть согласованы с компилятором moc из состава
      --  Qt. Для версии Qt 4.2.0 это src/tools/moc/generator.cpp.

      --  PropertyFlags
      --     Invalid = 0x00000000,
      --     Readable = 0x00000001,
      --     Writable = 0x00000002,
      --     Resettable = 0x00000004,
      --     EnumOrFlag = 0x00000008,
      --     StdCppSet = 0x00000100,
      --     Designable = 0x00001000,
      --     ResolveDesignable = 0x00002000,
      --     Scriptable = 0x00004000,
      --     ResolveScriptable = 0x00008000,
      --     Stored = 0x00010000,
      --     ResolveStored = 0x00020000,
      --     Editable = 0x00040000,
      --     ResolveEditable = 0x00080000,
      --     User = 0x00100000,
      --     ResolveUser = 0x00200000

      Invalid            : constant Flags_Type := 16#00000000#;
      Readable           : constant Flags_Type := 16#00000001#;
      Writable           : constant Flags_Type := 16#00000002#;
      Resettable         : constant Flags_Type := 16#00000004#;
      Enum_Or_Flag       : constant Flags_Type := 16#00000008#;
--    StdCppSet          : constant Flags_Type := 16#00000100#;
      Designable         : constant Flags_Type := 16#00001000#;
      Resolve_Designable : constant Flags_Type := 16#00002000#;
      Scriptable         : constant Flags_Type := 16#00004000#;
      Resolve_Scriptable : constant Flags_Type := 16#00008000#;
      Stored             : constant Flags_Type := 16#00010000#;
      Resolve_Stored     : constant Flags_Type := 16#00020000#;
      Editable           : constant Flags_Type := 16#00040000#;
      Resolve_Editable   : constant Flags_Type := 16#00080000#;
--    User               : constant Flags_Type := 16#00100000#;
--    Resolve_User       : constant Flags_Type := 16#00200000#;

      function Get_Type_Id (T : Unbounded_String) return Flags_Type;

      function Is_QVariant (T : Asis.Element) return Boolean;

      function Name_To_Type_Id (T : Asis.Element) return Flags_Type;

      -----------------
      -- Get_Type_Id --
      -----------------

      function Get_Type_Id (T : Unbounded_String) return Flags_Type is
      begin
         for J in Types_Array'Range loop
            if Ada.Characters.Handling.To_Upper
                (Ada.Strings.Unbounded.To_String (T))
               = Ada.Characters.Handling.To_Upper
                  (Ada.Strings.Unbounded.To_String
                   (Types_Array (J).Img))
            then
               return Flags_Type (Types_Array (J).Id);

            end if;
         end loop;

            return Flags_Type (257);
            --  XXX  Вместо 257 должен быть идентификатор типа, объявленного
            --       пользователем. Эквивалент того, что возвращает customType
            --       в qmetatype.cpp.
      end Get_Type_Id;

      -----------------
      -- Is_QVariant --
      -----------------

      function Is_QVariant (T : Asis.Element) return Boolean is
      begin
         if Name_To_Type_Id (T) /= 0 then
            return True;

         else
            return False;
         end if;
      end Is_QVariant;

      ---------------------
      -- Name_To_Type_Id --
      ---------------------

      function Name_To_Type_Id (T : Asis.Element) return Flags_Type is
         Result : Flags_Type := 0;

      begin
         if Asis.Elements.Is_Nil (T) then
            return 0;
         end if;

         declare
            Type_Image : constant Wide_String
              := Asis.Expressions.Name_Image (T);
            Image : constant String
              := Ada.Characters.Handling.To_Upper
                  (Ada.Characters.Conversions.To_String (Type_Image));

         begin
            if Image = "Q_VARIANT" then
               return 16#ffffffff#;
            end if;

            if Image = "Q_C_STRING" then
               return 12;
            end if;

            if Image = "Q_L_LONG" then
               return 4;
            end if;

            if Image = "Q_U_L_LONG" then
               return 5;
            end if;

            if Image = "Q_ICON_SET" then
               return 69;
            end if;

            if Image = "INT" then
               return 2;
            end if;

            Result := Get_Type_Id (To_Unbounded_String
                                   (Ada.Characters.Conversions.To_String
                                     (Type_Image)));

            if Result > 256 then
               return 0;
            end if;
         end;
            return Result;
      end Name_To_Type_Id;

      Flags : Flags_Type := Invalid;

      begin
         if Is_QVariant (Property.Prop_Type) then
            Flags := Flags xor Shift_Left
                                (Name_To_Type_Id (Property.Prop_Type), 24);

         else
            Flags := Flags xor Enum_Or_Flag;
         end if;

         if not (Asis.Elements.Is_Nil (Property.Read)) then
            Flags := Flags xor Readable;
         end if;

         if not (Asis.Elements.Is_Nil (Property.Write)) then
            Flags := Flags xor Writable;
         end if;

         if not (Asis.Elements.Is_Nil (Property.Reset)) then
            Flags := Flags xor Resettable;
         end if;

         if Property.Designable.Empty then
            Flags := Flags xor Resolve_Designable;
         elsif Property.Designable.State then
            Flags := Flags xor Designable;
         end if;

         if Property.Scriptable.Empty then
            Flags := Flags xor Resolve_Scriptable;
         elsif Property.Scriptable.State then
            Flags := Flags xor Scriptable;
         end if;

         if Property.Stored.Empty then
            Flags := Flags xor Resolve_Stored;
         elsif Property.Stored.State then
            Flags := Flags xor Stored;
         end if;

         if Property.Editable.Empty then
            Flags := Flags xor Resolve_Editable;
         elsif Property.Editable.State then
            Flags := Flags xor Editable;
         end if;

         return Flags;
      end Property_Flags;

      Spec_File_Name : constant String := Name & "-moc.ads";
      Body_File_Name : constant String := Name & "-moc.adb";
      Package_Name   : constant String
        := Ada.Characters.Conversions.To_String
            (Amoc.Model.Compilation_Unit_Name)
             & ".Moc";

   begin
      --  Generate package specification.

      Generated_Files.Append
       (Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
         (Ada.Characters.Conversions.To_Wide_String (Spec_File_Name)));

      Create (File, Out_File, Spec_File_Name);

      Put_Line (File, "pragma Style_Checks (""NM32766"");");

      New_Line (File);
      Put_Line (File, "with Interfaces.C;");
      Put_Line (File, "private with System;");

      New_Line (File);
      Put_Line (File, "with Qt4.Meta_Objects;");
      Put_Line (File, "with Qt_Ada.Internals.Object_Registry;");
      Put_Line (File, "private with Qt_Ada.Internals.Objects;");

      --  Add context clause for the parent class meta information package if
      --  this is required.

      if not Amoc.Model.Parent_Class_Is_Imported then
         New_Line (File);
         Put_Line
          (File,
           "with "
             & (-Amoc.Model.Parent_Package_Name)
             & ".Moc;");
      end if;

      New_Line (File);
      Put_Line (File, "package " & Package_Name & " is");
      New_Line (File);
      Put_Line (File, "   pragma Elaborate_Body;");

      New_Line (File);
      Put_Line (File, "   Static_Meta_Object :"
                        & " aliased constant Qt4.Meta_Objects.Q_Meta_Object;");

      Put_Line (File, "   Meta_Descriptor    : aliased constant");
      Put_Line
       (File,
        "     Qt_Ada.Internals.Object_Registry.Q_Object_Meta_Descriptor;");

      New_Line (File);
      Put_Line (File, "private");

      New_Line (File);
      Put_Line (File, "   pragma Warnings (Off);");
      Put_Line (File, "   use type Interfaces.C.char_array;");
      Put_Line (File, "   pragma Warnings (On);");

      New_Line (File);
      Put_Line (File, "   ------------------------");
      Put_Line (File, "   -- Static Meta Object --");
      Put_Line (File, "   ------------------------");
      New_Line (File);
      Generate_Data;
      New_Line (File);
      Generate_String_Data;

      New_Line (File);
      Put_Line (File, "   Static_Meta_Object :"
                        & " aliased constant Qt4.Meta_Objects.Q_Meta_Object");

      --  Generate reference to the parent class Static_Meta_Object object.

      if Amoc.Model.Parent_Class_Is_Imported then
         Put_Line
          (File,
           "     := ("
             & (-Amoc.Model.Parent_Package_Name)
             & ".Static_Meta_Object,");

      else
         Put_Line
          (File,
           "     := ("
             & (-Amoc.Model.Parent_Package_Name)
             & ".Moc.Static_Meta_Object'Access,");
      end if;

      Put_Line (File, "         String_Data (String_Data'First)'Access,");
      Put_Line (File, "         Data (Data'First)'Access,");
      Put_Line (File, "         null);");

      --  Generate meta descriptor.

      New_Line (File);
      Put_Line (File, "   ---------------------");
      Put_Line (File, "   -- Meta Descriptor --");
      Put_Line (File, "   ---------------------");
      New_Line (File);
      Put_Line (File, "   function Qt_Meta_Cast");
      Put_Line
       (File,
        "    (Self  : not null"
          & " Qt_Ada.Internals.Objects.Q_Object_Base_Access;");
      Put_Line (File, "     Class : String)");
      Put_Line (File, "       return System.Address;");
      New_Line (File);
      Put_Line (File, "   function Qt_Meta_Call");
      Put_Line
       (File,
        "    (Self : not null"
          & " Qt_Ada.Internals.Objects.Q_Object_Base_Access;");
      Put_Line (File, "     Kind : Qt4.Meta_Objects.Call;");
      Put_Line (File, "     Id   : Qt4.Q_Integer;");
      Put_Line (File, "     Args : Qt4.Address_Pointer)");
      Put_Line (File, "       return Qt4.Q_Integer;");
      New_Line (File);
      Put_Line (File, "   Meta_Descriptor : aliased constant");
      Put_Line
       (File,
        "     Qt_Ada.Internals.Object_Registry.Q_Object_Meta_Descriptor");

      if Amoc.Model.Parent_Class_Is_Imported then
         Put_Line (File, "       := (Parent      => null,");

      else
         Put_Line (File, "       := (Parent      =>");
         Put_Line
          (File,
           "             "
             & (-Amoc.Model.Parent_Package_Name)
             & ".Moc.Meta_Descriptor'Access,");
      end if;

      Put_Line
       (File, "           Meta_Object => Static_Meta_Object'Access,");
      Put_Line
       (File, "           Meta_Cast   => Qt_Meta_Cast'Access,");
      Put_Line
       (File, "           Meta_Call   => Qt_Meta_Call'Access);");

      New_Line (File);
      Put_Line (File, "end " & Package_Name & ";");
      Close (File);

      --  Generate package body.

      Generated_Files.Append
       (Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
         (Ada.Characters.Conversions.To_Wide_String (Body_File_Name)));

      Create (File, Out_File, Body_File_Name);

      Put_Line (File, "pragma Style_Checks (""NM32766"");");
      Put_Line (File, "pragma Warnings (Off);");
      New_Line (File);

      if Plugin.Is_Declared then
         Put_Line (File, "with Ada.Characters.Latin_1;");
      end if;

      Put_Line (File, "with System.Address_To_Access_Conversions;");
      New_Line (File);

      --  With required utilities packages for used QObject subclasses

      Amoc.Model.Signals.Iterate (Generate_Context_Clauses'Access);
      Amoc.Model.Slots.Iterate (Generate_Context_Clauses'Access);
      Amoc.Model.Methods.Iterate (Generate_Context_Clauses'Access);

      New_Line (File);
      Put_Line (File, "package body " & Package_Name & " is");

      --  Generate plugin verification data.

      if Plugin.Is_Declared then
         New_Line (File);
         Put_Line (File, "   Qt_Plugin_Verification_Data : constant "
                           & "Interfaces.C.char_array");

         for J in Amoc.Plugin_Verification.Data'Range loop
            Put_Line (File, Amoc.Plugin_Verification.Data (J).all);
         end loop;

         New_Line (File);
         Put_Line (File, "   function Qt_Plugin_Instance");
         Put_Line (File, "     return not null"
                           & " Qt_Ada.Internals.Objects.QObject_Access;");
         Put_Line (File, "   pragma Export (C,");
         Put_Line (File, "                  Qt_Plugin_Instance,");
         Put_Line (File, "                  ""qt_plugin_instance"");");
         New_Line (File);
         Put_Line (File, "   Instance :"
                           & " Qt_Ada.Internals.Objects.QObject_Access;");
         New_Line (File);
         Put_Line (File, "   function Qt_Plugin_Query_Verification_Data");
         Put_Line (File, "     return access constant Interfaces.C.char;");
         Put_Line (File, "   pragma Export (C,");
         Put_Line (File, "                  "
                           & "Qt_Plugin_Query_Verification_Data,");
         Put_Line (File, "                  "
                           & """qt_plugin_query_verification_data"");");
         New_Line (File);
         Put_Line (File, "   function Qt_Plugin_Instance");
         Put_Line (File, "     return not null"
                           & " Qt_Ada.Internals.Objects.QObject_Access");
         Put_Line (File, "   is");
         Put_Line (File, "      use type"
                           & " Qt_Ada.Internals.Objects.QObject_Access;");
         New_Line (File);
         Put_Line (File, "   begin");
         Put_Line (File, "      if Instance = null then");
         Put_Line (File, "         Instance :=");
         Put_Line (File, "           Qt_Ada.Internals.Objects.QObject_Access");
         Put_Line (File, "            ("
                           & Ada.Strings.Unbounded.To_String
                              (Plugin.Constructor)
                           & ".QObject_View);");
         Put_Line (File, "      end if;");
         New_Line (File);
         Put_Line (File, "      return Instance;");
         Put_Line (File, "   end Qt_Plugin_Instance;");
         New_Line (File);
         Put_Line (File, "   function Qt_Plugin_Query_Verification_Data");
         Put_Line (File, "     return access constant Interfaces.C.char");
         Put_Line (File, "   is");
         Put_Line (File, "   begin");
         Put_Line (File, "      return Qt_Plugin_Verification_Data");
         Put_Line (File, "              "
                           & "(Qt_Plugin_Verification_Data'First)'Access;");
         Put_Line (File, "   end Qt_Plugin_Query_Verification_Data;");
      end if;

      Generate_Meta_Call_Implementation (File);
      Generate_Meta_Cast_Implementation (File);

      New_Line (File);
      Put_Line (File, "begin");
      Put_Line (File, "   Qt_Ada.Internals.Object_Registry."
                        & "Register_QObject_Derived_Type");
      Put_Line
       (File,
        "    ("
          & (-Amoc.Model.Tagged_Type_Name)
          & "'Tag, Meta_Descriptor'Access);");
      Put_Line (File, "end " & Package_Name & ";");
      Close (File);
   end Generate_Meta_Data;

   -------------------------------------
   -- Generate_Signals_Implementation --
   -------------------------------------

   procedure Generate_Signals_Implementation is

      Signal_Number : Natural := 0;

      procedure Generate_Signal_Implementation
       (Position : Amoc.Model.Method_Info_Vectors.Cursor);
      --  Generates subunit with signal's implementation code.

      ------------------------------------
      -- Generate_Signal_Implementation --
      ------------------------------------

      procedure Generate_Signal_Implementation
       (Position : Amoc.Model.Method_Info_Vectors.Cursor)
      is
         File    : File_Type;
         Name    : constant String
           := Ada.Directories.Base_Name
               (Ada.Characters.Conversions.To_String
                 (Amoc.Parameters.Source_File_Name));
         Element : constant Amoc.Model.Method_Info
           := Amoc.Model.Method_Info_Vectors.Element (Position);
         Args    : constant Amoc.Model.Parameter_Vectors.Vector
           := Element.Parameters;

      begin
         Generated_Files.Append
          (Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String
            (Ada.Characters.Conversions.To_Wide_String
              (Name
                 & "-"
                 & Ada.Characters.Handling.To_Lower
                    (To_String (Element.Subprogram_Name))
                 & ".adb")));

         Ada.Text_IO.Create
          (File,
           Ada.Text_IO.Out_File,
           Name & '-'
             & Ada.Characters.Handling.To_Lower
                (To_String (Element.Subprogram_Name))
             & ".adb");
         Ada.Text_IO.Put_Line (File, "pragma Style_Checks (Off);");
         Ada.Text_IO.Put_Line (File, "pragma Warnings (Off);");
         Ada.Text_IO.Put_Line
          (File, "with System.Address_To_Access_Conversions;");
         Ada.Text_IO.Put_Line (File, "pragma Warnings (On);");
         Ada.Text_IO.New_Line (File);
         Ada.Text_IO.Put_Line (File, "with Qt4.Meta_Objects;");
         Ada.Text_IO.Put_Line (File, "with Qt4.Objects;");

         --  With required utilities packages for used QObject subclasses

         for K in 2 .. Integer (Args.Length) loop
            declare
               Param_Type : constant Amoc.Model.Type_Info
                 := Amoc.Model.Known_Types.Element
                     (Args.Element (K).Type_Id);

            begin
               if Param_Type.Kind = Amoc.Model.QObject_Subclass then
                  Ada.Text_IO.Put_Line
                   (File,
                    "with "
                      & Ada.Characters.Conversions.To_String
                         (Ada.Strings.Wide_Unbounded.To_Wide_String
                           (Param_Type.Utils_Package))
                      & ";");
               end if;
            end;
         end loop;

         Ada.Text_IO.New_Line (File);
         Ada.Text_IO.Put_Line
          (File,
           "with "
             & Ada.Characters.Conversions.To_String
                (Amoc.Model.Compilation_Unit_Name)
             & ".MOC;");
         Ada.Text_IO.New_Line (File);
         Ada.Text_IO.Put_Line
          (File,
           "separate ("
             & Ada.Characters.Conversions.To_String
                (Amoc.Model.Compilation_Unit_Name)
             & ')');

         Ada.Text_IO.Put_Line
          (File, "procedure " & To_String (Element.Subprogram_Name));

         for J in 1 .. Integer (Args.Length) loop
            if J = 1 then
               Ada.Text_IO.Put (File, " (");

            else
               Ada.Text_IO.Put (File, "  ");
            end if;

            Ada.Text_IO.Put
             (File,
              Ada.Characters.Conversions.To_String
               (Ada.Strings.Wide_Unbounded.To_Wide_String
                 (Args.Element (J).Name)));
            Ada.Text_IO.Put (File, " : ");

            if Args.Element (J).Is_Access then
               if Args.Element (J).Is_Not_Null then
                  Ada.Text_IO.Put (File, "not null ");
               end if;

               if Args.Element (J).Is_Constant then
                  Ada.Text_IO.Put (File, "access constant ");

               else
                  Ada.Text_IO.Put (File, "access ");
               end if;

            else
               if Args.Element (J).Is_Constant then
                  Ada.Text_IO.Put (File, "in ");

               else
                  raise Program_Error;
               end if;
            end if;

            if J = 1 then
               Ada.Text_IO.Put
                (File,
                 Ada.Characters.Conversions.To_String
                  (Ada.Strings.Wide_Unbounded.To_Wide_String
                    (Args.Element (J).Type_Image)));

            else
               Ada.Text_IO.Put
                (File,
                 Ada.Characters.Conversions.To_String
                  (Ada.Strings.Wide_Unbounded.To_Wide_String
                    (Amoc.Model.Known_Types.Element
                      (Args.Element (J).Type_Id).Ada_Name)));
            end if;

            if Args.Element (J).Is_Classwide then
               Ada.Text_IO.Put (File, "'Class");
            end if;

            if J = Integer (Args.Length) then
               Ada.Text_IO.Put_Line (File, ")");

            else
               Ada.Text_IO.Put_Line (File, ";");
            end if;
         end loop;

         Ada.Text_IO.Put_Line (File, "is");
         Ada.Text_IO.Put_Line (File, "   Data : aliased Qt4.Address_Array");
         Ada.Text_IO.Put (File, "     := (1 => System.Null_Address");

         for J in 2 .. Integer (Args.Length) loop
            declare
               Index      : constant String := Integer'Image (J);
               Image      : constant String
                 := Index (Index'First + 1 .. Index'Last);
               Param_Type : constant Amoc.Model.Type_Info
                 := Amoc.Model.Known_Types.Element
                     (Args.Element (J).Type_Id);

            begin
               Ada.Text_IO.Put_Line (File, ",");
               Ada.Text_IO.Put (File, "         " & Image & " => ");

               case Param_Type.Kind is
                  when Amoc.Model.Primitive
                    | Amoc.Model.Enumeration
                  =>
                     Ada.Text_IO.Put
                      (File,
                       Ada.Characters.Conversions.To_String
                        (Ada.Strings.Wide_Unbounded.To_Wide_String
                          (Args.Element (J).Name))
                         & "'Address");

                  when Amoc.Model.Value_Class =>
                     Ada.Text_IO.Put_Line
                      (File,
                       Ada.Characters.Conversions.To_String
                        (Ada.Strings.Wide_Unbounded.To_Wide_String
                          (Param_Type.To_Address)));
                     Ada.Text_IO.Put
                      (File,
                       "               ("
                         & Ada.Characters.Conversions.To_String
                            (Ada.Strings.Wide_Unbounded.To_Wide_String
                              (Param_Type.Class_View))
                         & " ("
                         & Ada.Characters.Conversions.To_String
                            (Ada.Strings.Wide_Unbounded.To_Wide_String
                              (Args.Element (J).Name))
                         & "))");

                  when Amoc.Model.QObject_Subclass =>
                     Ada.Text_IO.Put
                      (File,
                       Ada.Characters.Conversions.To_String
                        (Ada.Strings.Wide_Unbounded.To_Wide_String
                          (Param_Type.To_Address))
                         & " ("
                         & Ada.Characters.Conversions.To_String
                            (Ada.Strings.Wide_Unbounded.To_Wide_String
                              (Args.Element (J).Name))
                         & ")");
               end case;
            end;
         end loop;

         Ada.Text_IO.Put_Line (File, ");");
         Ada.Text_IO.New_Line (File);
         Ada.Text_IO.Put_Line (File, "begin");
         Ada.Text_IO.Put_Line (File, "   Qt4.Meta_Objects.Internals.Activate");

         declare
            Image : constant String := Natural'Image (Signal_Number);

         begin
            Ada.Text_IO.Put_Line
             (File,
              "    ("
                & Ada.Characters.Conversions.To_String
                   (Ada.Strings.Wide_Unbounded.To_Wide_String
                     (Args.First_Element.Name))
                & ",");

            Ada.Text_IO.Put_Line
             (File,
              "     "
                & Ada.Characters.Conversions.To_String
                   (Amoc.Model.Compilation_Unit_Name)
                & ".MOC.Static_Meta_Object'Access,");
            Ada.Text_IO.Put_Line
             (File, "     " & Image (Image'First + 1 .. Image'Last) & ",");
            Ada.Text_IO.Put_Line (File, "     Data (Data'First)'Address);");
         end;

         Ada.Text_IO.Put_Line
          (File, "end " & To_String (Element.Subprogram_Name) & ';');

         Ada.Text_IO.Close (File);

         Signal_Number := Signal_Number + 1;
      end Generate_Signal_Implementation;

   begin
      Amoc.Model.Signals.Iterate (Generate_Signal_Implementation'Access);
   end Generate_Signals_Implementation;

   -------------------------
   -- Get_Short_Type_Name --
   -------------------------

   function Get_Short_Type_Name (Full_Type : String) return String is
   begin
      for J in reverse Full_Type'Range loop
         case Full_Type (J) is
            when '.' =>
               return Full_Type (J + 1  .. Full_Type'Last);
            when others =>
               null;
         end case;
      end loop;

      return Full_Type;
   end Get_Short_Type_Name;

end Amoc.Generator;
