------------------------------------------------------------------------------
--                                                                          --
--                An Ada2005 binding to the Qt C++ framework                --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2008, 2009 Vadim Godunko <vgodunko@gmail.com>                --
--                                                                          --
-- QtAda is free software;  you can  redistribute it and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. QtAda is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License distributed with QtAda;  see file COPYING.  If not, write --
-- to  the  Free Software Foundation,  51  Franklin  Street,  Fifth  Floor, --
-- Boston, MA 02110-1301, USA.                                              --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 3056 $ $Date: 2009-01-08 23:27:35 +0300 (Thu, 08 Jan 2009) $
------------------------------------------------------------------------------
with Ada.Environment_Variables;
with Ada.Characters.Conversions;
with Ada.Command_Line;
with Ada.Strings.Wide_Fixed;
with Ada.Strings.Wide_Maps;
with Ada.Strings.Wide_Unbounded;
with Ada.Wide_Text_IO;

with Amoc.Version;

package body Amoc.Parameters is

   Amoc_Include_Path : constant String := "AMOC_INCLUDE_PATH";

   File_Name         : Ada.Strings.Wide_Unbounded.Unbounded_Wide_String;
   File_Name_Defined : Boolean := False;

   Source_Paths      : Wide_String_Vectors.Vector;

   Help_Flag         : Boolean := False;
   Version_Flag      : Boolean := False;
   Verbose_Flag      : Boolean := False;
   Dependencies_Flag : Boolean := False;

   function Path_Delimiter return Ada.Strings.Wide_Maps.Wide_Character_Set;
   --  Returns character set which include only one character, used on the
   --  platform as paths delimiter.

   ----------
   -- Help --
   ----------

   function Help return Boolean is
   begin
      return Help_Flag;
   end Help;

   -------------------------
   -- Output_Dependencies --
   -------------------------

   function Output_Dependencies return Boolean is
   begin
      return Dependencies_Flag;
   end Output_Dependencies;

   --------------------
   -- Path_Delimiter --
   --------------------

   function Path_Delimiter return Ada.Strings.Wide_Maps.Wide_Character_Set is
   begin
      case Amoc.Version.Qt_Ada_Platform is
         when Amoc.Version.Unix | Amoc.Version.MacOSX =>
            return Ada.Strings.Wide_Maps.To_Set (":");

         when Amoc.Version.Windows =>
            return Ada.Strings.Wide_Maps.To_Set (";");
      end case;
   end Path_Delimiter;

   ----------------------
   -- Source_File_Name --
   ----------------------

   function Source_File_Name return Wide_String is
   begin
      return
        Ada.Strings.Wide_Unbounded.To_Wide_String
         (File_Name);
   end Source_File_Name;

   ------------------------------
   -- Source_File_Search_Paths --
   ------------------------------

   function Source_File_Search_Paths return Wide_String_Vectors.Vector is
   begin
      return Source_Paths;
   end Source_File_Search_Paths;

   -------------
   -- Verbose --
   -------------

   function Verbose return Boolean is
   begin
      return Verbose_Flag;
   end Verbose;

   -------------
   -- Version --
   -------------

   function Version return Boolean is
   begin
      return Version_Flag;
   end Version;

begin
   if Ada.Command_Line.Argument_Count = 0 then
      Help_Flag := True;
   end if;

   for J in 1 .. Ada.Command_Line.Argument_Count loop
      declare
         Argument : constant Wide_String
           := Ada.Characters.Conversions.To_Wide_String
               (Ada.Command_Line.Argument (J));

      begin
         if Argument (Argument'First .. Argument'First + 1) = "--" then
            if Argument = "--help" then
               Help_Flag := True;

            elsif Argument = "--version" then
               Version_Flag := True;

            elsif Argument = "--verbose" then
               Verbose_Flag := True;

            else
               Ada.Wide_Text_IO.Put_Line
                ("Unhandled argument '" & Argument & "'");
            end if;

         elsif Argument (Argument'First) = '-' then
            case Argument (Argument'First + 1) is
               when 'I' =>
                  Source_Paths.Append
                   (Argument (Argument'First + 2 .. Argument'Last));

               when 'M' =>
                  Dependencies_Flag := True;

               when 'h' =>
                  Help_Flag := True;

               when 'v' =>
                  Verbose_Flag := True;

               when others =>
                  Ada.Wide_Text_IO.Put_Line
                   ("Unhandled argument '" & Argument & "'");
            end case;

         else
            if File_Name_Defined then
               raise Constraint_Error
                 with "only one source file must be specified";
            end if;

            File_Name :=
              Ada.Strings.Wide_Unbounded.To_Unbounded_Wide_String (Argument);
            File_Name_Defined := True;
         end if;
      end;
   end loop;

   --  Extract source search paths from the AMOC_INCLUDE_PATH environment
   --  variable.

   if Ada.Environment_Variables.Exists (Amoc_Include_Path) then
      declare
         Value : constant Wide_String
           := Ada.Characters.Conversions.To_Wide_String
               (Ada.Environment_Variables.Value (Amoc_Include_Path));
         First : Positive := Value'First;
         Last  : Natural;

      begin
         loop
            Ada.Strings.Wide_Fixed.Find_Token
             (Value (First .. Value'Last),
              Path_Delimiter,
              Ada.Strings.Outside,
              First,
              Last);

            exit when Last = 0;

            Source_Paths.Append (Value (First .. Last));

            First := Last + 1;
         end loop;
      end;
   end if;
end Amoc.Parameters;
